# -*- coding: utf-8 -*-
# Balazar safari photo
# Copyright (C) 2010-2014 Jean-Baptiste LAMY
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, struct, math, random, copy, os
from io import StringIO, BytesIO
import cerealizer
import PIL.Image

import soya3 as soya, soya3.gui, soya3.sdlconst as sdlconst

import balazar_safari_photo, balazar_safari_photo.globdef as globdef
import balazar_safari_photo.tofu.tofu4soya as tofu
from balazar_safari_photo.tofu.sides import *
from balazar_safari_photo.interface  import Bubble, LifeBar, close_bubble



ACTION_WAIT          = b" "
ACTION_LEFT          = b"<"
ACTION_LEFT_BACK     = b","
ACTION_LEFT_FRONT    = b"`"
ACTION_RIGHT         = b">"
ACTION_RIGHT_BACK    = b"."
ACTION_RIGHT_FRONT   = b"'"
ACTION_BACK          = b"_"
ACTION_FRONT         = b"^"
ACTION_ADVANCE       = b"A"
ACTION_ROLL          = b"R"
ACTION_JUMP          = b"j"
ACTION_STOP_JUMP     = b"J"
ACTION_PHOTO_BEGIN   = b"1"
ACTION_PHOTO_SHOOT   = b"2"
ACTION_PHOTO_END     = b"3"
ACTION_TURN_LEFT     = b"4"
ACTION_TURN_RIGHT    = b"5"
ACTION_GOTO          = b"g"
ACTION_GOTO_JUMP     = b"G"
ACTION_HURT          = b"H"
ACTION_DIE           = b"D"
ACTION_LOOK          = b"L"
ACTION_SHOUT         = b"S"
ACTION_STRIKE        = b"s"
ACTION_STRIKE2       = b"f"

ACTIONS_NON_CANCELABLE = set([ACTION_HURT, ACTION_DIE])

ACTIONS_NON_DIRECTION = set([ACTION_JUMP, ACTION_STOP_JUMP, ACTION_PHOTO_BEGIN, ACTION_PHOTO_SHOOT, ACTION_PHOTO_END, ACTION_HURT, ACTION_DIE, ACTION_LOOK, ACTION_SHOUT, ACTION_STRIKE, ACTION_STRIKE2])

ACTIONS_LEFT  = set([ACTION_LEFT , ACTION_LEFT_BACK , ACTION_LEFT_FRONT ])
ACTIONS_RIGHT = set([ACTION_RIGHT, ACTION_RIGHT_BACK, ACTION_RIGHT_FRONT])
ACTIONS_BACK  = set([ACTION_BACK , ACTION_LEFT_BACK , ACTION_RIGHT_BACK ])
ACTIONS_FRONT = set([ACTION_FRONT, ACTION_LEFT_FRONT, ACTION_RIGHT_FRONT])

ACTIONS_AUTHORIZED_WHILE_PHOTO = set([ACTION_WAIT, ACTION_TURN_LEFT, ACTION_TURN_RIGHT, ACTION_HURT, ACTION_DIE])

DIRECTIONS = { (0, 0, 1, 0) : ACTION_FRONT,
               (1, 0, 1, 0) : ACTION_LEFT_FRONT,
               (0, 1, 1, 0) : ACTION_RIGHT_FRONT,
               (1, 0, 0, 0) : ACTION_LEFT,
               (0, 1, 0, 0) : ACTION_RIGHT,
               (0, 0, 0, 1) : ACTION_BACK,
               (1, 0, 0, 1) : ACTION_LEFT_BACK,
               (0, 1, 0, 1) : ACTION_RIGHT_BACK,
               }

MESSAGE_LIFE      = b"L"
MESSAGE_BUBBLE    = b"B"
MESSAGE_QUEST     = b"Q"
MESSAGE_CHEST     = b"C"
MESSAGE_KEY       = b"K"
MESSAGE_ADD_POWER = b"P"


class BaseCamera(soya.Camera):
  def __init__(self, parent = None, target = None):
    soya.Camera.__init__(self, parent)
    self.photo_mode = 0
    self.set_target(target)
    
  def zap(self):
    # Ensure that the camera is the LAST children => it moves AFTER the character !
    parent = self.parent
    parent.remove(self)
    parent.add(self)
    
  def weather_changed(self):
    self.back = soya.MAIN_LOOP.weather.fog_end
    
  def is_point_visible(self, point, body, big = 0):
    p = point % self
    if p.z > 0.0: return 0
    if p.z < -self.back: return 0
    x, y = self.coord3d_to_2d(point)
    if (x < 0) or (y < 0) or (x > soya.get_screen_width()) or (y > soya.get_screen_height()): return 0
    vector.set_start_end(self, point)
    r = self.parent.raypick(self, vector, -1, 1, 1, raypick_collision, raypick_wall_normal)
    if big:
      if r and (raypick_collision.parent is body): return 1
      return 0
    else:
      if r:
        if raypick_collision.parent is body: return 1
        if self.distance_to(raypick_collision) < self.distance_to(point) - 0.01: return 0
      return 1
      
class Camera(BaseCamera):
  def set_target(self, target):
    self.target = target
    self.target_look_at = soya.Point(target, 0.0, 3.0, 0.0)
    
  def set_photo_mode(self, photo_mode):
    if photo_mode:
      self.photo_mode        = 1
      self.photo_mode_factor = 10.0
      self.photo_start_state = soya.CoordSystState(self)
    else:
      self.photo_mode_factor = -10.0
      self.photo_start_state = soya.CoordSystState(self)
    
  def advance_time(self, proportion):
    if self.photo_mode:
      if   self.photo_mode_factor > 0.0:
        self.photo_mode_factor = max(0.0, self.photo_mode_factor - 0.7 * proportion)
        f1 = self.photo_mode_factor / 10.0
        f2 = 1.0 - f1
        point.__init__(self.target, 0.0, 2.5, 2.0)
        point.convert_to(self.parent)
        self.set_xyz(f1 * self.photo_start_state.x + f2 * point.x, f1 * self.photo_start_state.y + f2 * point.y, f1 * self.photo_start_state.z + f2 * point.z)
        point.__init__(self.target, 0.0, f1 * 3.0 + f2 * 2.5, 0.0)
        self.look_at(point)
        
      elif self.photo_mode_factor < 0.0:
        self.photo_mode_factor = min(0.0, self.photo_mode_factor + 0.7 * proportion)
        if self.photo_mode_factor == 0.0: self.photo_mode = 0
        f1 = -self.photo_mode_factor / 10.0
        f2 = 1.0 - f1
        self.set_xyz(f1 * self.photo_start_state.x + f2 * self.target.x, f1 * self.photo_start_state.y + f2 * (self.target.y + 6.0), f1 * self.photo_start_state.z + f2 * (self.target.z + 15.0))
        point.__init__(self.target, 0.0, f1 * 2.5 + f2 * 3.0, 0.0)
        self.look_at(point)

      else:
        point.__init__(self.target, 0.0, 2.5, 2.0)
        self.move(point)
        self.look_at(self.target.front)

    else:
      self.set_xyz(self.target.x, self.target.y + 6.0, self.target.z + 15.0)
      self.look_at(self.target_look_at)
      
    soya.Camera.advance_time(self, proportion)
    
    
class FixedBgCamera(BaseCamera):
  def __init__(self, parent = None, target = None):
    BaseCamera.__init__(self, parent, target)
    
    self.min_x = int(0.4  * soya.get_screen_width ())
    self.max_x = int(0.6  * soya.get_screen_width ())
    self.min_y = int(0.07 * soya.get_screen_height())
    self.min_z =  4.0
    self.max_z = 35.0
    self.screen_width  = soya.get_screen_width ()
    self.screen_height = soya.get_screen_height()
    
    self.fixed_bg = soya.FixedBackgroundAtmosphere()
    self.no_bg    = soya.NoBackgroundNoDepthCleanAtmosphere()
    self.rendering_fixed_bg = 0
    self.y = 10000.0 # => Camera will be automatically repositionned and self.rendering_fixed_bg set to one
    
    self.center_on_wait = 1
    self.nb = 1
    
    self.weather_changed()
    porfkpero
    
  def set_target(self, target):
    self.target = target
    self.target_look_at = soya.Vector(None, 0.0, -3.0, -15.0)
    self.look_at(self.target_look_at)
    
  def set_photo_mode(self, photo_mode):
    if photo_mode == self.photo_mode: return
    if photo_mode:
      self.photo_mode = 1
      self.rendering_fixed_bg = 1
      point.__init__(self.target, 0.0, 2.5, 2.0)
      self.move(point)
      self.look_at(self.target.front)
      self.nb = 0.0
    else:
      self.photo_mode = 0
      self.look_at(self.target_look_at)
      self.rendering_fixed_bg = 1
      self.center(1, 1, 1)
      
  def advance_time(self, proportion):
    BaseCamera.advance_time(self, proportion)
    
    if soya.MAIN_LOOP.will_render:
      if self.photo_mode:
        if self.target.rotation_speed == 0.0:
          if self.nb > 5.0:
            point.__init__(self.target, 0.0, 2.5, 2.0)
            self.move(point)
            self.look_at(self.target.front)
            self.rendering_fixed_bg = 1
            self.nb = 0.0
        else: self.nb = 1.0
          
      else:
        point.clone(self.target)
        x, y = self.coord3d_to_2d(point)
        dz = self.z - self.target.z
        
        if self.center_on_wait:
          if self.target.speed.length() < 0.01:
            if (x < self.min_x) or (x > self.max_x) or (dz < 5.0) or (dz > 30.0):
              if self.nb > 5.0:
                self.center(1, 1, 1)
                self.rendering_fixed_bg = 1
                return
              
              elif self.nb: self.nb += proportion
          else: self.nb = 1.0
            
        if   x <= 0:
          self.center(1, 1, 1)
          if not self.center_on_wait: self.x -= 4.0
          self.rendering_fixed_bg = 1
        elif x >= self.screen_width:
          self.center(1, 1, 1)
          if not self.center_on_wait: self.x -= 4.0
          self.rendering_fixed_bg = 1
          
        if y < self.min_y:
          self.center(1, 1, 1)
          if self.target.current_animation == "chute": self.y += 2.0
          self.rendering_fixed_bg = 1
          
        elif y > self.screen_height:
          self.center(1, 1, 1)
          if self.target.current_animation == "chute": self.y -= 6.0
          self.rendering_fixed_bg = 1
          
        if not  self.center_on_wait:
          if dz < self.min_z:
            self.z = self.target.z + self.max_z - 5.0
            self.center(1, 1, 0)
            self.rendering_fixed_bg = 1
            
          elif dz > self.max_z:
            self.z = self.target.z + self.min_z + 7.0
            self.center(1, 1, 0)
            self.rendering_fixed_bg = 1
            
  def center(self, on_x, on_y, on_z):
    self.nb = 0.0
    if on_x: self.x = self.target.x
    if on_z: self.z = self.target.z + self.min_z + (self.max_z - self.min_z) * 0.5
    if on_y: self.y = self.target.y + 6.0 / 15.0 * (self.z - self.target.z)
    
  def weather_changed(self):
    BaseCamera.weather_changed(self)
    weather = soya.MAIN_LOOP.weather
    self.fixed_bg.fog       = self.no_bg.fog       = weather.fog
    self.fixed_bg.fog_start = self.no_bg.fog_start = weather.fog_start
    self.fixed_bg.fog_end   = self.no_bg.fog_end   = weather.fog_end
    self.fixed_bg.fog_color = self.no_bg.fog_color = weather.fog_color
    self.fixed_bg.bg_color  = self.no_bg.bg_color  = weather.bg_color
    self.fixed_bg.ambient   = self.no_bg.ambient   = weather.ambient
    
    self.rendering_fixed_bg = 1
    
  def render(self):
    if self.rendering_fixed_bg:
      saved_visibilities = {}
      for element in self.target.level.parent:
        if element is self.target.level: continue
        element.visible = not isinstance(element, Dynamic)
      for mobile in self.target.level.mobiles:
        saved_visibilities[mobile] = mobile.visible
        mobile.visible = not isinstance(mobile, Dynamic)
      self.target.level.static_world.visible = 1
      soya.MAIN_LOOP.weather.astral_world.visible = 1
      
      # Render fixed background
      soya.MAIN_LOOP.world.atmosphere = soya.MAIN_LOOP.weather
      soya.Camera.render(self)
      
      self.fixed_bg.capture() # Capture fixed background
      
      for element in self.target.level.parent:
        if element is self.target.level: continue
        element.visible = isinstance(element, Dynamic)
      for mobile in self.target.level.mobiles:
        mobile.visible = saved_visibilities[mobile]
      self.target.level.static_world.visible = 0
      soya.MAIN_LOOP.weather.astral_world.visible = 0
      
      # Render only mobiles
      soya.MAIN_LOOP.world.atmosphere = self.no_bg
      soya.Camera.render(self)
      
      if tofu.has_side("single"): soya.MAIN_LOOP.reset() # Do not count the time for capturing screen in FPS regulation
      
      # Future rendering would include restoring the fixed background + rendering mobiles
      soya.MAIN_LOOP.world.atmosphere = self.fixed_bg
      self.rendering_fixed_bg = 0
      
    else: soya.Camera.render(self)
    
    
JOY_BUTTON_KEYS = {
  0 : sdlconst.K_LSHIFT,
  1 : sdlconst.K_LCTRL,
  2 : sdlconst.K_ESCAPE,
  3 : sdlconst.K_4,
  4 : sdlconst.K_5,
  5 : sdlconst.K_6,
  6 : sdlconst.K_7,
  7 : sdlconst.K_8,
  }

class KeyboardController:
  def __init__(self, character):
    self.character = character
    self.left_key_down = self.right_key_down = self.up_key_down = self.down_key_down = self.control_down = 0
    self.last_click_x = self.last_click_y = -1000
    self.current_direction = ACTION_WAIT
    if globdef.FIXED_BG: self.camera = soya.MAIN_LOOP.camera = FixedBgCamera(soya.MAIN_LOOP.world, character)
    else:                self.camera = soya.MAIN_LOOP.camera = Camera(soya.MAIN_LOOP.world, character)
    
    layer = soya.gui.RootLayer(None)
    soya.gui.CameraViewport(layer, self.camera)
    soya.gui.FPSLabel(layer)
    soya.set_root_widget(layer)
    #character.life_bar = LifeBar()
    #character.life_bar.set_life(character.life)
    
  def destroy(self):
    self.camera.parent.remove(self.camera)
    
  def generate_actions(self):
    for event in soya.MAIN_LOOP.events:
      if   event[0] == sdlconst.JOYBUTTONDOWN: event = (sdlconst.KEYDOWN, JOY_BUTTON_KEYS.get(event[1]), "")
      elif event[0] == sdlconst.JOYBUTTONUP  :
        event = (sdlconst.KEYUP  , JOY_BUTTON_KEYS.get(event[1]), "")
        #print(event[1])
        
      if   event[0] == sdlconst.JOYAXISMOTION:
        #print(event)
        if   event[1] == 0:
          if   event[2] < 0: self.left_key_down  = 1
          elif event[2] > 0: self.right_key_down = 1
          else:              self.left_key_down  = self.right_key_down = 0
        elif event[1] == 1:
          if   event[2] < 0: self.up_key_down   = 1
          elif event[2] > 0: self.down_key_down = 1
          else:              self.up_key_down   = self.down_key_down = 0
          
      elif event[0] == sdlconst.KEYDOWN:
        if   event[1] == sdlconst.K_LEFT:   self.left_key_down  = 1
        elif event[1] == sdlconst.K_RIGHT:  self.right_key_down = 1
        elif event[1] == sdlconst.K_UP:     self.up_key_down    = 1
        elif event[1] == sdlconst.K_DOWN:   self.down_key_down  = 1
        
        elif event[1] == sdlconst.K_LSHIFT:
          close_bubble()
          if self.control_down:             self.character.send_action(ACTION_PHOTO_SHOOT); self.character.send_action(ACTION_PHOTO_END); self.control_down = 0
          else:                             self.character.send_action(ACTION_JUMP)
        elif event[1] == sdlconst.K_LCTRL:  self.character.send_action(ACTION_PHOTO_BEGIN); self.control_down = 1
          
        elif event[1] == sdlconst.K_ESCAPE:
          close_bubble()
          soya.MAIN_LOOP.stop(["i", self.character])
          
        elif event[1] == "q": soya.MAIN_LOOP.stop(["q"])
        
        elif event[1] == "t":
          print("    POSITION :  %s, %s, %s" % (self.character.x, self.character.y, self.character.z))
          self.character.set_life(self.character.life - 0.2)
          
      elif event[0] == sdlconst.KEYUP:
        if   event[1] == sdlconst.K_LSHIFT: self.character.send_action(ACTION_STOP_JUMP)
        elif event[1] == sdlconst.K_LCTRL:  self.character.send_action(ACTION_PHOTO_END); self.control_down = 0
        
        elif event[1] == sdlconst.K_LEFT:   self.left_key_down  = 0
        elif event[1] == sdlconst.K_RIGHT:  self.right_key_down = 0
        elif event[1] == sdlconst.K_UP:     self.up_key_down    = 0
        elif event[1] == sdlconst.K_DOWN:   self.down_key_down  = 0
        
      elif event[0] == sdlconst.MOUSEBUTTONDOWN:
        x = event[2]
        y = event[3]
        
        if   event[1] == 1:
          if abs(x - self.last_click_x) + abs(y - self.last_click_y) < 20:
            self.character.send_action(ACTION_JUMP)
          else:
            mouse = self.camera.coord2d_to_3d(x, y)
            vector.set_start_end(self.camera, mouse)
            r = self.camera.parent.raypick(self.camera, vector, -1, 1, 1, raypick_collision, raypick_wall_normal)
            if r:
              raypick_collision.convert_to(self.character.parent)
              self.character.send_action(ACTION_GOTO + struct.pack("!fff", raypick_collision.x, raypick_collision.y, raypick_collision.z))
          self.last_click_x = x
          self.last_click_y = y
        elif event[1] == 2:
          self.character.send_action(ACTION_PHOTO_BEGIN); self.control_down = 1
        elif event[1] == 3:
          if self.control_down:             self.character.send_action(ACTION_PHOTO_SHOOT); self.character.send_action(ACTION_PHOTO_END); self.control_down = 0
          else:                             self.character.send_action(ACTION_JUMP)
         
      elif event[0] == sdlconst.MOUSEBUTTONUP:
        if   event[1] == 2:
          self.character.send_action(ACTION_PHOTO_END); self.control_down = 0
        elif event[1] == 3:
          self.character.send_action(ACTION_STOP_JUMP)
          
    if self.control_down:
      if   self.left_key_down : new_direction = ACTION_TURN_LEFT
      elif self.right_key_down: new_direction = ACTION_TURN_RIGHT
      else:                     new_direction = ACTION_WAIT
    else:
      new_direction = DIRECTIONS.get((self.left_key_down, self.right_key_down, self.up_key_down, self.down_key_down), ACTION_WAIT)
      
    if new_direction != self.current_direction:
      self.current_direction = new_direction
      self.character.send_action(new_direction)



class Platform(Multisided):
  def _init(self):
    self.characters_on = set()
    
  def character_on(self, character):
    self.characters_on.add(character)
    return 1
    
  def character_off(self, character):
    self.characters_on.discard(character)
    return 1

  def remove_all_characters(self):
    for character in self.characters_on: character.platform = None
    self.characters_on = set()
    
    
raypick_collision   = soya.Point()
raypick_wall_normal = soya.Vector()
point               = soya.Point()
vector              = soya.Vector()
new_center          = soya.Point()
left                = soya.Vector(None, -1.0,  0.0,  0.0)
front               = soya.Vector(None,  0.0,  0.0, -1.0)
up                  = soya.Vector(None,  0.0,  1.0,  0.0)
down                = soya.Vector(None,  0.0, -1.0,  0.0)

class Dynamic(object): pass

class Hitable(Multisided):
  @side("single", "server")
  def set_life(self, life):
    if life <= 0.0: life = 0.0
    old_life = self.life
    if life < old_life:
      if self.current_action == ACTION_HURT: return # Cancel
      if life == 0.0: self.do_action(ACTION_DIE)
      else:           self.do_action(ACTION_HURT)
    self.send_message(MESSAGE_LIFE + struct.pack("!ff", life, self.current_rotation))
    self.hit()
    
  def hit(self):
    if self.life == 0.0: duration = 200.0
    else:                duration =  50.0
    perso = getattr(self, "perso", self)
    for mini_shader in perso.mini_shaders:
      if mini_shader.filename == "hurt":
        mini_shader.time = 0.0
        mini_shader._remove_at_time = duration
        mini_shader.duration        = duration
        break
    else:
      perso.add_mini_shader(soya.MiniShader.get("hurt")(remove_at_time = duration, color = [1.0, 0.0, 0.0, 1.0], duration = duration))
      
  def do_message(self, message):
    if   message[:1] == MESSAGE_LIFE:
      self.life, self.current_rotation = struct.unpack("!ff", message[1:])
      self.update_life()
    else:
      super(Hitable, self).do_message(message)
      
  def update_life(self): pass
  
  
class Character(tofu.Mobile, Dynamic):
  # We need RADIUS * sqrt(2)/2 < max speed (here, 0.35)
  RADIUS         = 1.0
  RADIUS_Y       = 1.0
  RADIUS_RAY     = max(RADIUS, 0.1 + RADIUS_Y)
  FLY            = 0
  JUMP_SPEED     = 0.5
  MOVE_SPEED     = 0.25
  ROTATION_SPEED = 6.0
  MIN_Y          = -9999.0
  TIME_BEFORE_KILLED = 200
  
  JUMP_SOUND = "jump1.wav"
  FALL_SOUND = "chute.wav"
  
  HURT_SOUND = "couic1.wav"
  DIE_SOUND  = "couic2.wav"
  
  def __init__(self, model_name = ""):
    tofu.Mobile.__init__(self)
    
    self.model_name       = model_name
    self.controller       = None
    self.life             = 1.0
    self.current_rotation = 0.0
    self.nb_fly           = 0
    self._init()
    
  def _init(self):
    if self.model_name:
      self.perso = soya.Body(self, soya.AnimatedModel.get(self.model_name))
      self.perso.animate_blend_cycle("attente")
      
    self.current_action    = ACTION_WAIT
    self.current_direction = ACTION_WAIT
    self.last_action       = ACTION_WAIT
    self.last_direction    = ACTION_WAIT
    self.current_animation = "attente"
    self.animation_nb      = 0
    
    self.solid = 0
    
    self.speed            = soya.Vector(soya.MAIN_LOOP and soya.MAIN_LOOP.scenes[0])
    self.rotation_speed   = 0.0
    self.center           = soya.Point(self, 0.0, self.RADIUS_Y, 0.0)
    self.platform         = None
    self.platform_pos     = soya.Point(None)
    self.platform_old_pos = soya.Point(None)
    
    self.left   = soya.Vector(self, -1.0,  0.0,  0.0)
    self.down   = soya.Vector(self,  0.0, -1.0,  0.0)
    self.up     = soya.Vector(self,  0.0,  1.0,  0.0)
    self.front  = soya.Vector(self,  0.0,  0.0, -1.0)
    
    self.goto_point = soya.Point(None)
    
    self.jumping   = 0
    self.on_ground = 0
    self.dec_x = self.dec_y = self.dec_z = self.dec_rotation = 0.0
    
  def loaded(self):
    tofu.Mobile.loaded(self)
    self._init()
    
  def __getstate__(self):
    return (self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.current_rotation, self.current_action, self.life, self.model_name, self.nb_fly, self.on_ground)
  
  def __setstate__(self, state):
    soya.World.__init__(self)
    self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.current_rotation, self.current_action, self.life, self.model_name, self.nb_fly, self.on_ground = state
    self.last_noticeable_round = 0
    self.last_important_round  = 0
    self.controller            = None
    
  def set_controller(self, controller):
    self.controller = controller
    if controller: self.controller.character = self
    
  def control_lost(self):
    tofu.Mobile.control_lost(self)
    self.set_controller(None)
    
  def do_message(self, message):
    if message[:1] == MESSAGE_BUBBLE:
      Bubble(message[1:].decode("utf8"), soya.Point(self, 0.0, 2.5, -0.5))
      
  def generate_actions(self):
    if self.controller: return self.controller.generate_actions()
    
  def play_sound(self, sound_name):
    if soya.MAIN_LOOP.camera:
      s = soya.SoundPlayer(soya.MAIN_LOOP.world, soya.Sound.get(sound_name), gain = 3.2 / (1.0 + self.distance_to(soya.MAIN_LOOP.camera)), play_in_3D = 0)
      s.move(soya.MAIN_LOOP.camera)
      
  def do_action(self, action):
    if (self.current_action in ACTIONS_NON_CANCELABLE): return
    
    if action[:1] in ACTIONS_NON_DIRECTION:
      if action != self.current_action:
        if   action == ACTION_HURT:
          self.current_direction = ACTION_WAIT
          self.animation_nb = 0
          Blood(self.get_root()).move(self.center)
          self.play_sound(self.HURT_SOUND)
        elif action == ACTION_DIE :
          self.current_direction = ACTION_WAIT
          self.animation_nb = 0
          self.start_die()
        elif action == ACTION_LOOK:
          self.current_direction = ACTION_WAIT
          self.animation_nb = 0
        elif action == ACTION_STRIKE:
          self.current_direction = ACTION_WAIT
          self.animation_nb = 0
        elif action == ACTION_STRIKE2:
          self.current_direction = ACTION_WAIT
          self.animation_nb = 0
        elif action == ACTION_SHOUT:
          self.current_direction = ACTION_WAIT
          self.animation_nb = 0
        self.current_action = self.last_action = action
        self.set_current_state_importance(2)
    else:
      if action != self.current_direction:
        self.current_direction = self.last_direction = action
        if (action[:1] == ACTION_GOTO) or (action[:1] == ACTION_GOTO_JUMP):
          self.goto_point.set_xyz(*struct.unpack("!fff", action[1:]))
        self.set_current_state_importance(2)
      
  def get_network_state(self):
    return self.last_action + self.last_direction + struct.pack("!ffff", self.x, self.y, self.z, self.current_rotation)
  
  def read_network_state(self, f):
    action    = f.read(1)
    direction = f.read(1)
    if (direction == ACTION_GOTO) or (direction == ACTION_GOTO_JUMP): direction += f.read(12)
    x, y, z, rotation = struct.unpack("!ffff", f.read(16))
    self.dec_x = x - self.x
    self.dec_y = y - self.y
    self.dec_z = z - self.z
    self.dec_rotation = rotation - self.current_rotation

    self.do_action(action)
    self.do_action(direction)
    
  def can_jump_before_falling(self): return 0
  
  def start_die(self):
    Blood(self.get_root(), 1).move(self.center)
    if self.DIE_SOUND: self.play_sound(self.DIE_SOUND)
    
  def do_physics(self):
    self.speed.x = self.speed.z = self.rotation_speed = 0.0
    
    if (not self.jumping) and self.speed.y > 0.0: self.speed.y = 0.0
    
    animation            = "attente"
    animation_prioritary = ""
    reverse_direction    = 0

    if   self.current_direction in ACTIONS_LEFT : self.speed.x -= self.MOVE_SPEED
    elif self.current_direction in ACTIONS_RIGHT: self.speed.x += self.MOVE_SPEED
    if   self.current_direction in ACTIONS_FRONT: self.speed.z -= self.MOVE_SPEED
    elif self.current_direction in ACTIONS_BACK : self.speed.z += self.MOVE_SPEED
    
    elif self.current_direction == ACTION_TURN_LEFT :
      self.rotation_speed =  self.ROTATION_SPEED
      animation = "tourneG"
      
    elif self.current_direction == ACTION_TURN_RIGHT:
      self.rotation_speed = -self.ROTATION_SPEED
      animation = "tourneD"
      
    elif self.current_direction == ACTION_ADVANCE:
      vector.__init__(self, 0.0, 0.0, -self.MOVE_SPEED)
      vector.convert_to(self.level)
      self.speed.x += vector.x
      self.speed.z += vector.z
      
    elif self.current_direction == ACTION_ROLL:
      if self.roll_speed:
        vector.set_xyz(self.dir_vector.x, 0.0, self.dir_vector.z)
        vector.set_length(min(self.MOVE_SPEED, 0.05 * self.roll_speed))
        self.speed.x += vector.x
        self.speed.z += vector.z
        
        self.roll_speed *= 0.95
        
      if self.roll_speed < 0.05:
        self.roll_speed = 0.0
        self.current_direction = ACTION_WAIT
        
    elif self.current_direction[:1] == ACTION_GOTO:
      self.speed.x = self.goto_point.x - self.x
      self.speed.z = self.goto_point.z - self.z
      if (abs(self.speed.x) < 0.01) and (abs(self.speed.z) < 0.01):
        self.current_direction = ACTION_WAIT
        self.speed.x = self.speed.z = 0.0
        
    elif self.current_direction[:1] == ACTION_GOTO_JUMP:
      self.speed.x = self.goto_point.x - self.x
      self.speed.z = self.goto_point.z - self.z
      if (abs(self.speed.x) < 0.01) and (abs(self.speed.z) < 0.01):
        self.speed.x = self.speed.z = 0.0
        if (self.nb_fly < self.FLY) and (abs(self.goto_point.y - self.y) > 0.1): pass
        else:
          self.current_direction = ACTION_WAIT
          
      if   self.goto_point.y > self.y + 1.0:
        self.current_action = ACTION_JUMP
        
    if   self.current_action == ACTION_HURT:
      animation_prioritary = "couic"
      reverse_direction    = 1
      vector.__init__(self, 0.0, 0.0, 0.2)
      vector.convert_to(self.speed.parent)
      self.speed.x = vector.x
      self.speed.z = vector.z
      if self.animation_nb == 0: self.speed.y = 0.3; self.jumping = 1
      if self.animation_nb > 15: self.current_action = ACTION_WAIT
      else:                      self.animation_nb += 1
      
    elif self.current_action == ACTION_DIE:
      animation_prioritary = "mort"
      reverse_direction    = 1
      if self.animation_nb > self.TIME_BEFORE_KILLED: self.kill()
      else:                                           self.animation_nb += 1
      
    elif self.current_action == ACTION_LOOK:
      animation_prioritary = "regarde"
      if self.animation_nb == self.LOOK_DURATION: self.look(); self.current_action = ACTION_WAIT
      else:                                       self.animation_nb += 1
      
    elif self.current_action == ACTION_SHOUT:
      animation_prioritary = "cri"
      if self.animation_nb == 5: self.shout()
      if self.animation_nb == self.SHOUT_DURATION: self.current_action = ACTION_WAIT
      else:                                        self.animation_nb += 1
      
    elif self.current_action == ACTION_STRIKE:
      animation_prioritary = "frappe"
      if self.animation_nb == self.STRIKE_DURATION: self.current_action = ACTION_WAIT
      else:                                         self.animation_nb += 1
      
    elif self.current_action == ACTION_STRIKE2:
      animation_prioritary = "frappe2"
      if self.animation_nb == self.STRIKE2_DURATION: self.current_action = ACTION_WAIT
      else:                                          self.animation_nb += 1
      
    if self.speed.x or self.speed.z:
      if self.speed.x and self.speed.z:
        if reverse_direction: vector.__init__(None, -self.speed.x, 0.0, -self.speed.z)
        else:                 vector.__init__(None,  self.speed.x, 0.0,  self.speed.z)
        if vector.length() > 0.25:
          vector.set_length(0.25)
          self.speed.x = vector.x
          self.speed.z = vector.z
          
      if reverse_direction: angle = (math.atan2( self.speed.x,  self.speed.z) * 57.295779513082323)
      else:                 angle = (math.atan2(-self.speed.x, -self.speed.z) * 57.295779513082323)
      delta = (self.current_rotation - angle) % 360.0
      if    10.0 <  delta < 180.0: animation = "tourneD"; self.rotation_speed = -self.ROTATION_SPEED * 1.67
      elif 180.0 <= delta < 350.0: animation = "tourneG"; self.rotation_speed =  self.ROTATION_SPEED * 1.67
      else:
        animation = "marche"
        self.rotation_speed = (angle - self.current_rotation) % 360.0
        if self.rotation_speed > 180.0: self.rotation_speed -= 360.0
        
        
    self.speed.add_xyz(self.dec_x, 0.0, self.dec_z)
    self.rotation_speed += self.dec_rotation
    self.dec_x = self.dec_y = self.dec_z = self.dec_rotation = 0.0
    
    
    saved_solid = self.solid
    self.solid  = 0
    
    new_center.__init__(self, 0.0, self.RADIUS_Y, 0.0)
    new_center.add_vector(self.speed)
    
    context = soya.MAIN_LOOP.scenes[0].RaypickContext(new_center, self.RADIUS_RAY)
    
    r = context.raypick(new_center, self.down, 0.2 + self.RADIUS_Y, 1, 1, raypick_collision, raypick_wall_normal)
    if self.y - 0.2 < self.MIN_Y:
      r = 1
      raypick_collision.__init__(self.level, self.x, self.MIN_Y, self.z)
    if r and not self.jumping: # Put character on ground
      self.nb_fly    = 0
      self.on_ground = 1
      parent = raypick_collision.parent
      while parent and not isinstance(parent, tofu.Mobile):
        parent = parent.parent
      if parent: # On a mobile platform
        self.platform_pos    .clone(raypick_collision)
        self.platform_old_pos.clone(raypick_collision)
        self.platform_old_pos.convert_to(self.parent)
        if not self.platform is parent:
          if self.platform: self.platform.character_off(self)
          if parent.character_on(self):
            self.platform = parent
      else: # On a static element
        if self.platform:
          self.platform.character_off(self)
          self.platform = None
          
      if not self.jumping: # platform may have changed the jumping status !
        if self.current_action == ACTION_JUMP:
          self.jumping        = 1
          self.speed.y        = self.JUMP_SPEED
          self.current_action = ACTION_WAIT
          self.nb_fly         = 1
          self.play_sound(self.JUMP_SOUND)
          
        else:
          raypick_collision.convert_to(self)
          if (self.speed.y < -0.2) and (raypick_collision.y): self.play_sound(self.FALL_SOUND)
          self.speed.y = raypick_collision.y

          
    else: # Falling
      if self.platform:
        self.platform.character_off(self)
        self.platform = None
      if self.on_ground:
        if self.can_jump_before_falling(): self.current_action = ACTION_JUMP
        self.on_ground = 0
        
      if (self.current_action == ACTION_JUMP) and (self.nb_fly < self.FLY):
        self.jumping        = 1
        if self.nb_fly: self.speed.y        = self.JUMP_SPEED * 0.7
        else:           self.speed.y        = self.JUMP_SPEED
        self.current_action = ACTION_WAIT
        self.nb_fly         += 1
      else:
        self.speed.y = max(self.speed.y - 0.03, -0.25)
        animation = "chute"
        
        if self.current_action == ACTION_STOP_JUMP:
          self.current_action = ACTION_WAIT
          if self.speed.y < self.JUMP_SPEED:
            if self.speed.y > 0.0: self.speed.y = 0.0
            self.jumping = 0
        elif self.speed.y < 0.0: self.jumping = 0
        
    if (self.speed.x != 0) or (self.speed.z != 0) or (self.speed.y > 0.0):
      new_center.y += self.speed.y
      
      r = context.raypick(new_center, self.left, self.RADIUS, 0, 1, raypick_collision, raypick_wall_normal)
      if r:
        hypo = self.RADIUS - new_center.distance_to(raypick_collision)
        raypick_wall_normal.__imul__(hypo)
        raypick_wall_normal.convert_to(self.parent)
        raypick_wall_normal.y = 0.0
        self.speed.add_vector(raypick_wall_normal)
        new_center.add_vector(raypick_wall_normal)
        
      r = context.raypick(new_center, self.front, self.RADIUS, 0, 1, raypick_collision, raypick_wall_normal)
      if r:
        hypo = self.RADIUS - new_center.distance_to(raypick_collision)
        raypick_wall_normal.__imul__(hypo)
        raypick_wall_normal.convert_to(self.parent)
        raypick_wall_normal.y = 0.0
        self.speed.add_vector(raypick_wall_normal)
        new_center.add_vector(raypick_wall_normal)
        
      r = context.raypick_b(new_center, self.up, self.RADIUS, 1, 1, raypick_collision, raypick_wall_normal)
      if r:
        if self.speed.y > 0.0: self.speed.y = 0.0
        
    self.play_animation(animation_prioritary or animation)
    self.solid = saved_solid
    
  def play_animation(self, animation):
    if self.current_animation != animation:
      self.perso.animate_clear_cycle(self.current_animation, 0.2)
      self.perso.animate_reset()
      self.perso.animate_blend_cycle(animation, 1.0, 0.2)
      self.current_animation = animation
      
  def update_platform_pos(self):
    vector.set_start_end(self.platform_old_pos, self.platform_pos)
    self.add_vector(vector)
    self.platform_old_pos.clone(self.platform_pos)
    self.platform_old_pos.convert_to(self.parent)
    
  def advance_time(self, proportion):
    tofu.Mobile.advance_time(self, proportion)
    
    x, y, z = self.x, self.y, self.z
    self.set_identity()
    self.x, self.y, self.z = x, y, z
    self.current_rotation = (self.current_rotation + proportion * self.rotation_speed) % 360.0
    self.rotate_y(self.current_rotation)
    self.add_mul_vector(proportion, self.speed)
    
    if self.platform: self.update_platform_pos()
    

class Monster(Character):
  def get_photo_name(self): return None


class PhotoVortex(soya.PythonCoordSyst, Dynamic):
  def __init__(self, parent = None):
    soya.PythonCoordSyst.__init__(self, parent)
    self.state    = 0.0
    self.material = soya.Material.get("fx_photo")
    
  def destroy(self):
    if self.parent: self.parent.remove(self)
    
  def advance_time(self, proportion):
    if self.state > 75.0:
      soya.IDLER.next_round_tasks.append(self.destroy)
    else:
      self.state += proportion * 2.0
      
  def batch(self):
    return 2, self, self.material
  
  def render(self):
    opengl = soya.opengl
    
    self.material.activate()
    
    opengl.glDisable(opengl.GL_CULL_FACE)
    opengl.glDisable(opengl.GL_LIGHTING)
    opengl.glDisable(opengl.GL_FOG)
    opengl.glBegin(opengl.GL_TRIANGLES)
    
    angle = (-10.0) * 0.017453292519943295
    ac    = math.cos(angle)
    az    = math.sin(angle)
    h     = math.sin(self.state * 0.062831853071795868) * 10.0
    
    color = 1
    for i in range(1):
      for angle in range(0, 360, 10):
        angle = (angle) * 0.017453292519943295
        c = math.cos(angle)
        s = math.sin(angle)
        z1 = - i      * 15.0 - min(0, h)
        z2 = -(i + 1) * 15.0 - h
        
        opengl.glColor4f(0.3, 0.7, 1.0, 1.0 - self.state / 75.0)
        opengl.glTexCoord2f(0.5, 1.0); opengl.glVertex3f(c  * z1, s  * z1, z1)
        
        opengl.glColor4f(1.0, 1.0, 1.0, 0.0)
        opengl.glTexCoord2f(1.0, 0.04); opengl.glVertex3f(ac * z2, az * z2, z2)
        opengl.glTexCoord2f(0.0, 0.04); opengl.glVertex3f(c  * z2, s  * z2, z2)
        
        color = not color
        ac = c
        az = s
        
    opengl.glEnd()
    opengl.glColor4f(1.0, 1.0, 1.0, 1.0)
    opengl.glEnable(opengl.GL_CULL_FACE)
    opengl.glEnable(opengl.GL_LIGHTING)
    opengl.glEnable(opengl.GL_FOG)
    
    
class Blood(soya.Particles):
  def __init__(self, parent = None, big = 0):
    soya.Particles.__init__(self, parent, soya.PARTICLE_DEFAULT_MATERIAL, 40)
    self.auto_generate_particle = 1
    self.set_colors((0.2, 0.0, 0.0, 1.0), (0.2, 0.04, 0.04, 1.0), (0.2, 0.01, 0.01, 1.0), (0.0, 0.0, 0.0, 1.0))
    self.set_sizes((0.75, 0.75))
    self.lit  = 0
    self.particle_coordsyst = soya.MAIN_LOOP.world
    if big:
      self.max_particles_per_round = 5
      self.life = 90
    else:
      self.max_particles_per_round = 4
      self.life = 15
      
  def generate(self, index):
    angle = random.uniform(0.0, 6.2834)
    sx = math.cos(angle)
    sz = math.sin(angle)
    l = 0.05 / math.sqrt(sx * sx + 0.2 + sz * sz)
    self.set_particle(index, random.uniform(0.5, 1.5), sx * l, 5 * l, sz * l, 0.0, -0.015, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    if self.life == 1:
      self.auto_generate_particle = 0
      self.removable = 1
      self.life = 0
    else: self.life -= 1


class EjectedPiece(soya.Body, Dynamic):
  def __init__(self, character, model_name):
    soya.Body.__init__(self, character.level.parent, soya.Model.get(model_name))
    self.move(character.perso)
    
    self.solid    = 0
    self.rot_lat  = random.uniform(-10.0, 10.0)
    self.rot_vert = random.uniform(-10.0, 10.0)
    self.speed    = soya.Vector(None, random.uniform(-1.0, 1.0), random.uniform(1.0, 3.0), random.uniform(-1.0, 1.0))
    self.speed.set_length(0.4)
    
  def advance_time(self, proportion):
    soya.Body.advance_time(self, proportion)
    self.rotate_lateral (proportion * self.rot_lat)
    self.rotate_vertical(proportion * self.rot_vert)
    self.add_mul_vector(proportion, self.speed)
    
  def begin_round(self):
    soya.Body.begin_round(self)
    self.speed.y -= 0.008
    if self.y < -40.0: self.parent.remove(self)
    
cerealizer.register(EjectedPiece)


    
def new_photo(character, monsters, image):
  if isinstance(image, str):
    jpeg = open(image, "rb").read()
    image = PIL.Image.open(image)
  else:
    jpeg = BytesIO()
    image.save(jpeg, "jpeg")
    jpeg = jpeg.getvalue()
    
  image = image.resize((128, 128), 1)
  mini_jpeg = BytesIO()
  image.save(mini_jpeg, "jpeg")
  mini_jpeg = mini_jpeg.getvalue()
  
  quests  = list(QUESTS.values())
  quests.sort(key = lambda quest: quest.priority)
  already = {}
  for photo in character.photos:
    if photo.quest_name: already[photo.quest_name] = photo
  best1 = best2 = None
  for quest in quests:
    if quest.satisfy(monsters):
      if quest.name in already:
        if not best2: best2 = quest
      else:
        if not best1: best1 = quest
  best = best1 or best2
  if best: quest_name = best.name
  else:    quest_name = ""
  
  if character.photos: id = character.photos[-1].id + 1
  else:                id = 0
  
  #open("/tmp/t2.jpeg", "w").write(jpeg)
  
  return Photo(id, character.level.filename, monsters, quest_name, jpeg, mini_jpeg)

class Photographiable(object):
  def get_photo_name(self): return None
  
class BasePhoto(object):
  def get_message(self):
    if not self.monsters: return _("__empty_photo__")
    monsters        = [_(monster) for monster in self.monsters]
    monsters_labels = []
    for monster in set(monsters):
      nb = monsters.count(monster)
      if nb == 1: monsters_labels.append(_("1 %s" % monster))
      else:
        translation = _("x %s" % monster)
        if "%s" in translation: monsters_labels.append(translation % nb)
        else:
          monsters_labels.append("%s %s" % (nb, translation))
          print("missing translation", "x %s" % monster)
    if   len(monsters_labels) == 1: monsters_label = monsters_labels[0]
    else:                           monsters_label = ",".join(monsters_labels[:-1]) + _("__and_photo__") + monsters_labels[-1]
    if len(monster) == 1: return _("__monster_photo__")  % monsters_label
    else:                 return _("__monsters_photo__") % monsters_label
    
  def destroy(self): pass


QUESTS = {}
class Quest(object):
  def __init__(self, priority, name, text, level_name, photo_x, photo_y, photo_angle, satisfy):
    self.priority    = priority
    self.name        = name
    self.text        = text
    self.level_name  = level_name
    self.photo_x     = photo_x
    self.photo_y     = photo_y
    self.photo_angle = photo_angle
    self.satisfy     = satisfy
    QUESTS[name] = self
  
  def get_label(self):
    labels = []
    for i in self.name.split("\n"):
      if i.startswith("1"): labels.append(_(i))
      else:                 labels.append(_("x%s" % i[1:]) % i[0])
    return " +\n".join(labels)
    
POWERS = {}
class Power(object):
  def __init__(self, name, quest_names):
    self.name        = name
    self.quest_names = quest_names
    POWERS[name] = self


class Photo(BasePhoto):
  def __init__(self, id, level_name, monsters, quest_name, jpeg, mini_jpeg):
    self.id         = id
    self.level_name = level_name
    self.monsters   = monsters
    self.quest_name = quest_name
    self.jpeg       = jpeg
    self.mini_jpeg  = mini_jpeg
    
  def __getstate__(self):
    return self.id, self.level_name, self.monsters, self.quest_name, self.jpeg, self.mini_jpeg
  
  def __setstate__(self, state):
    self.id, self.level_name, self.monsters, self.quest_name, self.jpeg, self.mini_jpeg = state
    
  def get_material(self):
    image = PIL.Image.open(BytesIO(self.jpeg))
    width, height = image.size
    if   width  <  512: width  =  512
    elif width  < 1024: width  = 1024
    elif width  < 2048: width  = 2048
    elif width  < 4096: width  = 4096
    if   height <  512: height =  512
    elif height < 1024: height = 1024
    elif height < 2048: height = 2048
    elif height < 4096: height = 4096
    image = image.resize((width, height), 0)
    material = soya.Material(soya.image_from_pil(image))
    material.clamp = 1
    return material
    
  def get_mini_material(self):
    material = soya.Material(soya.image_from_pil(PIL.Image.open(BytesIO(self.mini_jpeg))))
    material.clamp = 1
    return material

  def get_saved_photo(self, player): return SavedPhoto(self, player)
    
class SavedPhoto(BasePhoto):
  def __init__(self, photo, player):
    self.id         = photo.id
    self.level_name = photo.level_name
    self.monsters   = photo.monsters
    self.quest_name = photo.quest_name
    
    self.filename = "%s_%s_%s" % (player.filename, self.level_name, self.id)
    #self.filename = os.path.join(tofu.Player.DIRNAME, "%s_%s_%s" % (player.filename, self.level_name, self.id)
    
    open(os.path.join(balazar_safari_photo.tofu.path[0], tofu.Player.DIRNAME, "%s.jpeg"      % self.filename), "wb").write(photo.jpeg)
    open(os.path.join(balazar_safari_photo.tofu.path[0], tofu.Player.DIRNAME, "mini_%s.jpeg" % self.filename), "wb").write(photo.mini_jpeg)
    
  def destroy(self):
    os.unlink(os.path.join(balazar_safari_photo.tofu.path[0], tofu.Player.DIRNAME, "%s.jpeg"      % self.filename))
    os.unlink(os.path.join(balazar_safari_photo.tofu.path[0], tofu.Player.DIRNAME, "mini_%s.jpeg" % self.filename))
    
  def __getstate__(self):
    return self.id, self.level_name, self.monsters, self.quest_name, self.filename
  
  def __setstate__(self, state):
    self.id, self.level_name, self.monsters, self.quest_name, self.filename = state

  def get_photo(self):
    return Photo(self.id, self.level_name, self.monsters, self.quest_name, open(os.path.join(balazar_safari_photo.tofu.path[0], tofu.Player.DIRNAME, "%s.jpeg" % self.filename), "rb").read(), open(os.path.join(balazar_safari_photo.tofu.path[0], tofu.Player.DIRNAME, "mini_%s.jpeg" % self.filename), "rb").read())
  
    
class Balazar(Hitable, Character):
  RADIUS     = 1.0
  RADIUS_Y   = 1.0
  RADIUS_RAY = max(RADIUS, 0.1 + RADIUS_Y)
  
  
  FLY        = 1
  life_bar   = None
  instruction_nb = 0
  
  def __init__(self):
    Character.__init__(self, "balazar")
    #Character.__init__(self, "monopode")
    #Character.__init__(self, "caraconque")
    #Character.__init__(self, "echassien")
    #Character.__init__(self, "pretre")
    self.photos = []
    self.chests = set()
    self.powers = set()
    self.keys   = {}
    self.min_y  = -20.0
    
  def init_interface(self):
    self.level_name = self.level.filename # For inventory
    
  def _init(self):
    Character._init(self)
    self.shadow_effect = soya.Body(self, soya.Model.get("ombre"))
    self.shadow_effect.solid = 0
    
    self.right_hand = soya.World(self)
    self.perso.attach_to_bone(self.right_hand, "mainD")
    
    self.photo_cam = soya.Body(self.right_hand, soya.Model.get("camera"))
    self.photo_cam.rotate_lateral(180.0)
    self.photo_cam.rotate_vertical(90.0)
    self.photo_cam.set_xyz(0.02, 0.0, -0.4)
    self.photo_cam.add_mini_shader(soya.MiniShader.get("camera")(color = [0.5, 0.7, 0.8, 1.0]))
    
    self.photo_pose = 0
    self.last_ground_pos = soya.Point(self.level, self.x, self.y, self.z)
    
    if self.life <= 0.0: self.do_action(ACTION_DIE)
    
  def __getstate__(self):
    return (Character.__getstate__(self), self.chests, self.keys, self.photos, self.powers, self.min_y)
  
  def __setstate__(self, state):
    super_state, self.chests, self.keys, self.photos, self.powers, self.min_y = state
    Character.__setstate__(self, super_state)
    
    if "plume" in self.powers:
      self.FLY = 2
      
  def update_life(self):
    if self.life_bar: self.life_bar.set_life(self.life)
    
  def killed(self):
    soya.MAIN_LOOP.stop(["r"])
    
  def control_owned(self):
    tofu.Mobile.control_owned(self)
    self.set_controller(KeyboardController(self))
    self.set_chests_visibility()
    
  def set_chests_visibility(self):
    for i in self.level.mobiles:
      if isinstance(i, balazar_safari_photo.bonus.SmallChest):
        i.visible = not (self.level.filename, i.chest_id) in self.chests
        
  def start_die(self):
    Character.start_die(self)
    for i in range(self.keys.get(self.level.filename, 0)):
      p = EjectedPiece(self, "clef")
      
  def control_lost(self):
    tofu.Mobile.control_lost(self)
    self.controller.destroy()
    self.set_controller(None)
    
  def do_action(self, action):
    if (self.current_action in ACTIONS_NON_CANCELABLE): return
    
    if   action == ACTION_PHOTO_BEGIN:
      self.current_direction = self.last_direction = ACTION_WAIT
      self.current_action    = self.last_action    = ACTION_PHOTO_BEGIN
      if self.local:
        self.controller.camera.set_photo_mode(1)
        #if globdef.QUALITY == 2: self.photo_cam.add_deform(self.photo_cam_deform)
      self.set_current_state_importance(2)
      
    elif action == ACTION_PHOTO_END:
      self.last_action    = ACTION_PHOTO_END
      self.current_action = ACTION_WAIT
      if self.photo_pose:
        self.photo_pose = 0
        if self.local:
          self.controller.camera.set_photo_mode(0)
          #if globdef.QUALITY == 2: self.photo_cam.remove_deform(self.photo_cam_deform)
        self.set_current_state_importance(2)
        
    elif action == ACTION_PHOTO_SHOOT:
      if self.photo_pose > 15:
        self.last_action = ACTION_PHOTO_SHOOT
        self.set_current_state_importance(2)
        
        if self.local:
          hiddens = [self]
          if self.life_bar: hiddens.append(self.life_bar)
          for widget in soya.root_widget.widgets:
            if isinstance(widget, soya.gui.FPSLabel):
              widget.text = ""
              hiddens.append(widget)
          for i in hiddens: i.visible = 0
          soya.render()
          soya.render() # 2 rendering due to double / triple buffering
          image    = soya.screenshot(use_back_buffer = 1)
          monsters = list(filter(None, [monster.get_photo_name() for monster in self.level.mobiles + self.level.static_world.children if isinstance(monster, Photographiable)]))
          photo    = new_photo(self, monsters, image)
          extra    = ""
          
          if photo.quest_name:
            for i in range(len(self.photos)):
              if self.photos[i].quest_name == photo.quest_name:
                self.photos[i].destroy()
                del self.photos[i]
                break
            else:
              if photo.quest_name == "Monopode":
                extra += "\n" + _("__tuto_4__")
                
            quest_name_2_photo =  { other_photo.quest_name : other_photo for other_photo in self.photos if other_photo.quest_name }
            quest_name_2_photo[photo.quest_name] = photo
            
            for power in POWERS.values():
              if photo.quest_name in power.quest_names:
                for quest_name in power.quest_names:
                  if not quest_name in quest_name_2_photo: break
                else:
                  self.send_message(MESSAGE_ADD_POWER + power.name.encode("utf8"))
                  extra += "\n" + _("__power_%s__" % power.name)
                  
          else:
            photos_without_quest = [i for i in self.photos if not i.quest_name]
            if len(photos_without_quest) > 6:
              photos_without_quest[0].destroy()
              self.photos.remove(photos_without_quest[0])
              
          self.photos.append(photo)
          
          for i in hiddens: i.visible = 1
          
          self.send_message(MESSAGE_BUBBLE + (photo.get_message() + extra).encode("utf8"))
          
          if tofu.has_side("single"):
            soya.MAIN_LOOP.reset() # Do not count the time for capturing screen in FPS regulation
            
        vortex = PhotoVortex()
        self.level.parent.add(vortex)
        vortex.move(self.photo_cam)
        vector.__init__(self.photo_cam, 1.0, 0.0, 0.0)
        vortex.look_at(vector)
        
    elif (self.photo_pose == 0) or (action in ACTIONS_AUTHORIZED_WHILE_PHOTO):
      Character.do_action(self, action)
      
  def do_message(self, message):
    if   message[:1] == MESSAGE_CHEST:
      chest_id = int(message[1:].decode("utf8"))
      self.chests.add((self.level.filename, chest_id))
      
    elif message[:1] == MESSAGE_KEY:
      self.keys[self.level.filename] = self.keys.get(self.level.filename, 0) + 1
      
    elif message[:1] == MESSAGE_ADD_POWER:
      power = message[1:].decode("utf8")
      self.powers.add(power)
      if power == "plume":
        if self.FLY < 2: self.FLY = 2
        
    else: super(Balazar, self).do_message(message)
    
  def do_physics(self):
    Character.do_physics(self)
    if self.current_action == ACTION_PHOTO_BEGIN:
      self.photo_pose += 1
      
    if 0.0 < self.life < 1.0:
      self.life += 0.002
      if self.life_bar: self.life_bar.set_life(self.life)
      
      
  def do_collisions(self):
    if self.y < self.min_y:
      self.set_life(self.life - 0.4)
      if self.life > 0: self.move(self.last_ground_pos)
      
    else:
      if self.on_ground and (not self.platform) and (self.y > -17.0):
        self.last_ground_pos.move(self)
        
  def play_animation(self, animation):
    if self.photo_pose and (animation != "couic") and (animation != "mort"): animation = "photographie"
    Character.play_animation(self, animation)
    
  def advance_time(self, proportion):
    Character.advance_time(self, proportion)
    
    if soya.MAIN_LOOP.will_render:
      r = self.parent.raypick(self.center, self.down, -1, 1, 1, raypick_collision, raypick_wall_normal)
      if r:
        raypick_collision.convert_to(self)
        self.shadow_effect.y = raypick_collision.y + 0.02
        self.shadow_effect.look_at_y(raypick_wall_normal)
      else:
        self.shadow_effect.y = -9999.0
        
cerealizer.register(Photo)
cerealizer.register(SavedPhoto)
cerealizer.register(Balazar)


