# -*- coding: utf-8 -*-
# Balazar safari photo
# Copyright (C) 2010-2014 Jean-Baptiste LAMY
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, struct, math, random, copy, os
import cerealizer

import soya3 as soya, soya3.sdlconst as sdlconst

import balazar_safari_photo, balazar_safari_photo.globdef as globdef
import balazar_safari_photo.tofu.tofu4soya as tofu
from balazar_safari_photo.tofu.sides import *
from balazar_safari_photo.base       import *

class Bonus(tofu.Mobile):
  TAKEN_SOUND = "ouais.wav"
  
  def __init__(self):
    tofu.Mobile.__init__(self)
    
    self.bot       = 1
    self.radius    = 2.0
    self.animation = 0
    self.bonus_id  = 0
    self._init()
    
  def _init(self):
    self.solid = 0
    self.sprite = soya.Sprite(self, soya.Material.get("x_lumiere_1"))
    self.sprite.y = 1.5
    self.sprite.width = self.sprite.height = 2.0
    self.sprite.lit = 0
    
  def character_on(self, character): return 0
  def character_off(self, character): pass
  
  def added_into_level(self, level):
    if level: self.bonus_id = len([i for i in level if isinstance(i, Bonus)])
    
  def loaded(self):
    tofu.Mobile.loaded(self)
    self._init()
    
  def __getstate__(self):
    return (self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.bonus_id, self.animation)
  
  def __setstate__(self, state):
    soya.World.__init__(self)
    self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.bonus_id, self.animation = state
    self.radius    = 1.0
    
  def do_collisions(self):
    tofu.Mobile.do_collisions(self)
    
    if self.visible and not self.animation:
      for character in self.level.mobiles:
        if isinstance(character, Balazar) and (character.center.distance_to(self) < self.radius + character.RADIUS):
          self.taken_by(character)
          
  def taken_by(self, character):
    if character.local:
      self.animation = 1
      character.play_sound(self.TAKEN_SOUND)
      
  def resurrected(self):
    self.set_scale_factors(1.0, 1.0, 1.0)
    self.visible   = 1
    self.animation = 0
    
  def advance_time(self, proportion):
    if self.visible:
      if self.animation:
        self.rotate_lateral((0.2 * self.animation + 4.0) * proportion)
        
        self.animation += proportion
        
        if   self.animation <  50.0: self.y += 0.15 * proportion
        elif self.animation <  80.0: pass
        elif self.animation < 100.0: self.scale(0.9, 0.9, 0.9); self.sprite.color = [1.0, 1.0, 1.0, self.sprite.color[3] * 0.9]
        else:                        self.hide()
          
      else: self.rotate_lateral(4.0 * proportion)
  
  def hide(self):
    self.visible   = 0
    self.animation = 0
    self.y -= 0.15 * 50.0
    

class SmallChest(Bonus, Dynamic):
  def __init__(self, chest_id, key_speed = [-0.027, 0.165, 0.11]):
    Bonus.__init__(self)
    self.chest_id = chest_id
    self.key_speed = key_speed
    
  def _init(self):
    Bonus._init(self)
    
    self.model = soya.Model.get("coffre1@typ1")
    self.lid   = soya.Body(self, soya.Model.get("coffre_couvercle1@typ1"))
    self.lid.set_xyz(0.0, 2.0, 1.0)
    self.scale(0.5, 0.5, 0.5)
        
  def __getstate__(self):
    return (Bonus.__getstate__(self), self.chest_id, self.key_speed)
  
  def __setstate__(self, state):
    super_state, self.chest_id, self.key_speed = state
    Bonus.__setstate__(self, super_state)
    
  @side("single", "client")
  def taken_by(self, character):
    Bonus.taken_by(self, character)
    character.send_message(MESSAGE_CHEST + str(self.chest_id).encode("utf8"))
    #tofu.Player.get(character.player_name).chests_found[self.level.filename, self.bonus_id] = 1
    
    k = Key(self.key_speed)
    k.set_xyz(self.x, self.y, self.z)
    self.level.add_mobile(k)
    
  def resurrected(self):
    Bonus.resurrected(self)
    self.set_scale_factors(0.5, 0.5, 0.5)
    self.lid.set_identity(); self.lid.set_xyz(0.0, 2.0, 1.0)
    
  def advance_time(self, proportion):
    Bonus.advance_time(self, proportion)
    if self.visible and self.animation: self.lid.rotate_vertical(3.0 * proportion)

cerealizer.register(SmallChest)

class Key(Bonus, Dynamic):
  def __init__(self, speed):
    Bonus.__init__(self)
    
    self.speed = soya.Vector(None, *speed)
    
  def _init(self):
    Bonus._init(self)
    self.model = soya.Model.get("clef")
    
  def __getstate__(self):
    return (Bonus.__getstate__(self), self.speed)
  
  def __setstate__(self, state):
    super_state, self.speed = state
    Bonus.__setstate__(self, super_state)
    
  def advance_time(self, proportion):
    Bonus.advance_time(self, proportion)
    if self.speed: self.add_mul_vector(proportion, self.speed)
    
  def begin_round(self):
    Bonus.begin_round(self)
    if self.speed: self.speed.y -= 0.009
    
  def do_collisions(self):
    if self.speed:
      tofu.Mobile.do_collisions(self) # Do not do normal bonus collision until on the ground!
      
      if self.speed.y < 0.0:
        r = soya.MAIN_LOOP.scenes[0].raypick(self, down, 0.1 - self.speed.y, 1, 1, raypick_collision, raypick_wall_normal)
        if r:
          self.move(raypick_collision)
          self.speed = None
    else:
      Bonus.do_collisions(self)
      
  @side("single", "client")
  def taken_by(self, character):
    Bonus.taken_by(self, character)
    character.send_message(MESSAGE_KEY)

  def hide(self): self.kill()

cerealizer.register(Key)



#Quest(5, "Champignassien" , "une photo d'un Champignassien"            , "champi1",   0, 484,  2.0, lambda monsters: "Champignassien"   in monsters)
Quest(5, "1 Champignassien"  , "une photo d'un Champignassien"            , "champi1", 110, 218, -1.0, lambda monsters: "Champignassien"   in monsters)
Quest(7, "1 Monopode"        , "une photo d'un Monopode"                  , "champi1", 110, 312,  3.0, lambda monsters: ("Monopode" in monsters) or ("ShoutingMonopode" in monsters))
Quest(3, "1 ShoutingMonopode", "une photo d'un Monopode en train de crier", "champi1", 106, 407, -2.0, lambda monsters: "ShoutingMonopode" in monsters)
Quest(1, "2 Monopode"        , "une photo de deux Monopodes"              , "champi1", 110, 499,  4.0, lambda monsters: monsters.count("Monopode") >= 2)
Quest(2, "1 BabyMonopode"    , "une photo d'un bébé monopode"             , "champi1", 242, 498, -3.0, lambda monsters: "BabyMonopode" in monsters)

Quest(7, "1 Bolet\n1 Caraconque", "une photo d'une tortue-limace et d'un bolet" , "champi2", 373, 498,  2.0, lambda monsters: ("Caraconque" in monsters) and ("Bolet" in monsters))
Quest(5, "1 Echassien"          , "une photo d'un Échassien"                    , "champi2", 503, 499, -2.0, lambda monsters: "Echassien" in monsters)
Quest(5, "1 EchassienPriest"    , "une photo d'un prêtre Échassien en gros plan", "champi2", 634, 497, 3.0, lambda monsters: "EchassienPriest" in monsters)
Quest(3, "2 Bolet"              , "une photo de deux Bolets"                    , "champi2", 766, 498, -3.0, lambda monsters: monsters.count("Bolet") >= 2)

Power("plume", ["Champignassien", "Monopode", "Monopode criant", "deux Monopodes", "bébé Monopode"])
#Power("plume", ["Champignassien", "Monopode"])

