# -*- coding: utf-8 -*-
# Balazar safari photo
# Copyright (C) 2010-2014 Jean-Baptiste LAMY
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, struct, math, random, copy, os
import cerealizer
from types import GeneratorType

import soya3 as soya

import balazar_safari_photo, balazar_safari_photo.globdef as globdef
import balazar_safari_photo.tofu.tofu4soya as tofu
from balazar_safari_photo.tofu.sides import *
from balazar_safari_photo.base       import *

AI_MODE_NORMAL  = 1
AI_MODE_ALERTED = 2


class KillPoint(soya.Point):
  def __init__(self, parent, x, y, z, radius, damage):
    soya.Point.__init__(self, parent, x, y, z)
    self.radius = radius
    self.damage = damage


class AIControlled(Multisided):
  kill_point = None
  
  @side("single", "server")
  def _init(self):
    self.ai_generators  = [self.ai()]
    self.ai_last_result = None
    
  def generate_actions(self):
    #if not self.character.level: return
    while 1:
      try:
        r = self.ai_generators[-1].send(self.ai_last_result)
        if   isinstance(r, GeneratorType): # A nested child generator
          self.ai_generators.append(r)
          self.ai_last_result = None # Required for new generator
        elif not r is None: # A return value
          self.ai_last_result = r
          del self.ai_generators[-1]
        return
      
      except StopIteration:
        del self.ai_generators[-1]

  def nearest_balazar(self):
    best      = None
    best_dist = 99999.0
    for i in self.level.mobiles:
      if isinstance(i, Balazar) and (i.life > 0.0):
        dist = self.distance_to(i)
        if (not best) or (dist < best_dist):
          best      = i
          best_dist = dist
    return best
        
  def look(self):
    for i in self.level.mobiles:
      if isinstance(i, Balazar) and (self.distance_to(i) < self.LOOK_DISTANCE):
        vector.set_start_end(self, i)
        if (self.LOOK_ANGLE >= 180.0) or (self.front.angle_to(vector) < self.LOOK_ANGLE):
          self.alerted(i, 0)
          return 1
        
  def shout(self):
    for i in self.level.mobiles:
      if isinstance(i, AIControlled):
        i.alerted(None, 1)
        
  def alerted(self, balazar, by_shout = 0):
    self.ai_mode = AI_MODE_ALERTED
    
  def ai(self):
    yield None
  
  def ai_goto(self, x, y, z, jump = 1):
    last_pos = soya.Point()
    last_pos.clone(self)
    if jump: self.send_action(ACTION_GOTO_JUMP + struct.pack("!fff", x, y, z))
    else:    self.send_action(ACTION_GOTO      + struct.pack("!fff", x, y, z))
    while 1:
      yield
      if self.current_direction == ACTION_WAIT: yield 1
      if last_pos.distance_to(self) < 0.06:
        yield 0
        break
      last_pos.clone(self)
      
  def ai_look_toward(self, toward):
    while 1:
      vector.set_start_end(self, toward)
      vector.y = 0.0
      if vector.angle_to(self.front) > 30.0:
        point.clone(toward)
        point.convert_to(self)
        if point.x < 0:
          if self.current_direction != ACTION_TURN_LEFT : self.send_action(ACTION_TURN_LEFT)
        else:
          if self.current_direction != ACTION_TURN_RIGHT: self.send_action(ACTION_TURN_RIGHT)
        yield
      else:
        if self.current_direction != ACTION_WAIT: self.send_action(ACTION_WAIT)
        break
      
  def ai_wait(self, nb_round):
    for i in range(nb_round): yield
    yield 1
    
  def ai_look(self):
    self.send_action(ACTION_LOOK)
    for i in range(self.LOOK_DURATION + 1): yield
    yield 1
    
  def ai_shout(self):
    self.send_action(ACTION_SHOUT)
    for i in range(self.SHOUT_DURATION): yield
    yield 1
    
  def ai_strike(self):
    self.send_action(ACTION_STRIKE)
    for i in range(self.STRIKE_DURATION): yield
    yield 1
    
  def ai_strike2(self):
    self.send_action(ACTION_STRIKE2)
    for i in range(self.STRIKE2_DURATION): yield
    yield 1
    
    
class Striking(object):
  def do_collisions(self):
    super(Striking, self).do_collisions()
    
    if self.kill_point and ((self.current_action == ACTION_STRIKE) or (self.current_action == ACTION_STRIKE2)) and (self.animation_nb > 5):
      for mobile in self.level.mobiles:
        if isinstance(mobile, Hitable):
          if mobile.center.distance_to(self.kill_point) < mobile.RADIUS + self.kill_point.radius:
            mobile.current_rotation = (self.current_rotation + 180.0) % 360.0
            mobile.set_life(mobile.life - self.kill_point.damage)
            
      
class Champignassien(AIControlled, Platform, Monster, Photographiable):
  # We need RADIUS * sqrt(2)/2 < max speed
  RADIUS        = 3.0
  RADIUS_Y      = 6.0
  RADIUS_RAY    = max(RADIUS, 0.1 + RADIUS_Y)
  FLY           = 99999
  FLY_Y = 10.0
  
  JUMP_SOUND = "morfle1.wav"
  FALL_SOUND = "morfle1.wav"


  def __init__(self):
    Monster     .__init__(self, "champignassien")
    AIControlled.__init__(self)
    
  def _init(self):
    self.nb_attente = 0
    AIControlled._init(self)
    Platform    ._init(self)
    Monster     ._init(self)
    self.solid = 1
    
  def __getstate__(self):
    return (Monster.__getstate__(self))
  
  def __setstate__(self, state):
    super_state = state
    Monster.__setstate__(self, super_state)
    
  def get_photo_name(self):
    if not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 10.5, -0.2), self.perso): return None
    vector.__init__(soya.MAIN_LOOP.camera, 0.0, 0.0, -1.0)
    if self.front.dot_product(vector) > -0.3: return None
    return self.__class__.__name__
  
  def begin_round(self):
    Monster.begin_round(self)
    
    if self.current_animation == "attente":
      self.nb_attente += 1
      if self.nb_attente == 100:
        self.perso.model = soya.Model.get("champignassien@pos1")
    else:
      self.nb_attente  = 0
      
  def play_animation(self, animation):
    if (animation == "marche") or (animation == "tourneD") or (animation == "tourneG"): animation = "chute"
    if not isinstance(self.perso.model, soya.AnimatedModel):
      if (animation != "attente"): self.perso.model = soya.AnimatedModel.get("champignassien")
      else:                        return
      
    Monster.play_animation(self, animation)
    
  def ai(self):
    yield self.ai_pose()

    nb = random.randint(4000, 5000)
    while 1:
      
      yield
      nb -= 1
      for mobile in self.level.mobiles:
        if not mobile.bot:
          if mobile.distance_to(self) < 5.0:
            nb -= 50
            break
      if (nb <= 0) or self.characters_on:
        yield self.ai_envole()
        for i in range(random.randint(50, 100)):
          if self.y < self.FLY_Y:
            self.send_action(ACTION_JUMP)
            yield
            self.send_action(ACTION_WAIT)
          yield
        yield self.ai_pose()
        nb = random.randint(500, 1500)
        
  def ai_envole(self):
    point.__init__(self, 0.0, 0.0, -2.0)
    point.convert_to(self.parent)
    x0 = self.x
    z0 = self.z
    r = yield self.ai_goto(point.x, self.FLY_Y, point.z)
    if not r:
      r = yield self.ai_goto(x0, self.FLY_Y, z0)
      if not r:
        yield self.ai_goto(x0, self.y, z0)
        yield self.ai_goto(x0, self.FLY_Y, z0)
        
    self.send_action(ACTION_WAIT)
    
  def ai_pose(self):
    mushrooms = [element for element in self.level.static_world if isinstance(element, soya.Body) and element.model and (element.model.filename == "champignon")]
    while 1:
      yield
      mushroom = random.choice(mushrooms)
      #mushroom = soya.Point(self.level, 50.0, 0.0, 25.0)
      vector.__init__(None, random.uniform(-1.0, 1.0), 0.0, random.uniform(-1.0, 1.0))
      if (vector.x == 0.0) and (vector.z == 0.0): continue
      vector.set_length(12.0)
      x = mushroom.x + vector.x
      z = mushroom.z + vector.z
      point.__init__(None, x, self.y, z)
      soya.MAIN_LOOP.scenes[0].raypick(point, down, -1, 1, 1, raypick_collision, raypick_wall_normal)
      y0 = raypick_collision.y
      point.x += 2.0
      if soya.MAIN_LOOP.scenes[0].raypick(point, down, -1, 1, 1, raypick_collision, raypick_wall_normal):
        if raypick_collision.y > y0 + 5.0: continue
      point.x -= 4.0
      if soya.MAIN_LOOP.scenes[0].raypick(point, down, -1, 1, 1, raypick_collision, raypick_wall_normal):
        if raypick_collision.y > y0 + 5.0: continue
      point.x += 2.0; point.z += 2.0
      if soya.MAIN_LOOP.scenes[0].raypick(point, down, -1, 1, 1, raypick_collision, raypick_wall_normal):
        if raypick_collision.y > y0 + 5.0: continue
      point.z -= 4.0
      if soya.MAIN_LOOP.scenes[0].raypick(point, down, -1, 1, 1, raypick_collision, raypick_wall_normal):
        if raypick_collision.y > y0 + 5.0: continue
        
      break
    
    r = yield self.ai_goto(x, self.FLY_Y, z)
    
    self.send_action(ACTION_WAIT)
    while 1:
      yield
      if abs(self.speed.y) < 0.0001: break
        
cerealizer.register(Champignassien)


class Monopode(AIControlled, Platform, Monster, Photographiable):
  # We need RADIUS * sqrt(2)/2 < max speed
  RADIUS         = 2.0
  RADIUS_Y       = 1.4
  RADIUS_RAY     = max(RADIUS, 0.1 + RADIUS_Y)
  JUMP_SPEED     = 0.7
  MOVE_SPEED     = 0.3
  LOOK_DURATION  = 40
  LOOK_DISTANCE  = 15.0
  LOOK_ANGLE     = 80.0
  SHOUT_DURATION = 18
  
  CALM_DISTANCE  = 30.0
  
  def __init__(self):
    Monster     .__init__(self, "monopode")
    AIControlled.__init__(self)
    
  def _init(self):
    self.ai_mode = AI_MODE_NORMAL
    AIControlled._init(self)
    Platform    ._init(self)
    Monster     ._init(self)
    self.solid = 1
    
  def __getstate__(self): return (Monster.__getstate__(self), self.ai_mode)
  
  def __setstate__(self, state):
    super_state, self.ai_mode = state
    Monster.__setstate__(self, super_state)
    
  def can_jump_before_falling(self):
    point.__init__(self, 0.0, 0.0, -1.0)
    r = self.level.raypick(point, down, -1.0, 1, 1, raypick_collision, raypick_wall_normal)
    if r:
      raypick_collision.convert_to(self.level)
      if self.y - raypick_collision.y < 2.0:
        return 0
    return 1
  
  def get_photo_name(self):
    if self.distance_to(soya.MAIN_LOOP.camera) > 30.0: return None
    if not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 1.5, -0.2), self.perso): return None
    if self.current_action == ACTION_SHOUT: return "Shouting%s" % self.__class__.__name__
    return self.__class__.__name__
  
  def ai(self):
    while 1:
      if   self.ai_mode == AI_MODE_NORMAL:
        point.__init__(self.level, self.x + random.uniform(-12.0, 12.0), self.y + 3.0, self.z + random.uniform(-12.0, 12.0))
        self.solid = 0
        r = self.level.raypick(point, down, -1.0, 1, 1, raypick_collision, raypick_wall_normal)
        self.solid = 1
        if r:
          raypick_collision  .convert_to(self.level)
          raypick_wall_normal.convert_to(self.level)
          if (raypick_collision.y > -16.0) and (raypick_collision.y - self.y > -3.0) and (raypick_wall_normal.y > 0.5):
            yield self.ai_goto(raypick_collision.x, raypick_collision.y, raypick_collision.z)
            
        yield self.ai_wait(150)
        
        yield self.ai_look()
        if self.ai_mode == AI_MODE_ALERTED:
          yield self.ai_shout()
          
      elif self.ai_mode == AI_MODE_ALERTED:
        balazar = self.nearest_balazar()
        if (not balazar) or (self.distance_to(balazar) > self.CALM_DISTANCE):
          self.ai_mode = AI_MODE_NORMAL
          continue
        
        for i in range(7):
          if i < 5:
            vector.set_start_end(balazar, self)
            vector.set_length(random.uniform(5.0, 10.0))
          else:
            vector.set_start_end(self, balazar)
            vector.__imul__(1.1)
          vector.convert_to(self.level)
          point.__init__(self.level, self.x + vector.x + random.uniform(-1.0, 1.0), self.y + 3.0, self.z + vector.z + random.uniform(-1.0, 1.0))
          self.solid = 0
          r = self.level.raypick(point, down, -1.0, 1, 1, raypick_collision, raypick_wall_normal)
          self.solid = 1
          if r:
            raypick_collision  .convert_to(self.level)
            raypick_wall_normal.convert_to(self.level)
            if (raypick_collision.y > -17.5) and (raypick_collision.y - self.y > -3.0) and (raypick_wall_normal.y > 0.5):
              raypick_collision.convert_to(self.level)
              yield self.ai_goto(raypick_collision.x, raypick_collision.y, raypick_collision.z)
              break
            
        else:
          yield self.ai_wait(10)
    
    
cerealizer.register(Monopode)


class BabyMonopode(Monopode):
  # We need RADIUS * sqrt(2)/2 < max speed
  RADIUS         = 1.3
  RADIUS_Y       = 1.3
  RADIUS_RAY     = max(RADIUS, 0.1 + RADIUS_Y)
  JUMP_SPEED     = 1.0
  MOVE_SPEED     = 0.3
  LOOK_DURATION  = 40
  LOOK_DISTANCE  = 30.0
  LOOK_ANGLE     = 120.0
  SHOUT_DURATION = 9
  
  CALM_DISTANCE  = 50.0
  
  def __init__(self):
    Monster     .__init__(self, "bebe_monopode")
    AIControlled.__init__(self)
      
  def get_photo_name(self):
    if self.distance_to(soya.MAIN_LOOP.camera) > 30.0: return None
    if not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 1.5, -0.2), self.perso): return None
    return self.__class__.__name__
    
    
    
cerealizer.register(BabyMonopode)


class Egg(Platform, Monster, Photographiable, Hitable):
  # We need RADIUS * sqrt(2)/2 < max speed
  RADIUS          = 2.2
  RADIUS_Y        = 2.2
  RADIUS_RAY      = max(RADIUS, 0.1 + RADIUS_Y)
  JUMP_SPEED      = 0.3
  MOVE_SPEED      = 0.1
  MIN_Y           = -19.0
  ROTATION_SPEED  = 2.0
  TIME_BEFORE_KILLED = 0
  DIE_SOUND          = ""
    
  def __init__(self):
    Monster.__init__(self, "")
    
  def _init(self):
    Platform._init(self)
    Monster ._init(self)
    self.perso = soya.Body(self, soya.Model.get("oeuf"))
    self.perso.add_mini_shader(soya.MiniShader.get("egg")())
    self.perso.y = 1.7
    self.solid = 1
    self.current_action = ACTION_ROLL
    self.roll_speed  = 0.0
    self.roll_vector = None
    self.dir_vector  = soya.Vector()
    self.life = 0.05
    
  def get_photo_name(self):
    if self.distance_to(soya.MAIN_LOOP.camera) > 40.0: return None
    if (not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 0.5, 0.0), self.perso)): return None
    return self.__class__.__name__
    
  def generate_actions(self):
    if self.characters_on:
      character = tuple(self.characters_on)[0]
      vector.set_start_end(self, character)
      vector.convert_to(self.level)
      self.roll_vector = vector.cross_product(up)
      vector.y = 0.0
      self.roll_speed = vector.length()
      
      self.dir_vector.set_xyz(vector.x, 0.0, vector.z)
      self.do_action(ACTION_ROLL)
      
  def do_physics(self):      
    Monster.do_physics(self)
    
  def advance_time(self, proportion):
    Monster.advance_time(self, proportion)
    
    if self.roll_speed:
      self.perso.rotate_axis(-self.roll_speed * proportion, self.roll_vector)

  def play_animation(self, animation):
    if self.current_animation != animation: self.current_animation = animation
    
  def start_die(self):
    for i in range(30): p = EjectedPiece(self, "oeuf_eclat")
    
  def killed(self):
    b = BabyMonopode(); b.set_xyz(self.x, self.y, self.z); self.level.add_mobile(b)
    
    super().killed()
    
cerealizer.register(Egg)




class Caraconque(AIControlled, Platform, Monster, Photographiable):
  # We need RADIUS * sqrt(2)/2 < max speed
  RADIUS          = 4.0
  RADIUS_Y        = 1.4
  RADIUS_RAY      = max(RADIUS, 0.1 + RADIUS_Y)
  JUMP_SPEED      = 0.3
  MOVE_SPEED      = 0.2
  MIN_Y           = -19.0
  ROTATION_SPEED  = 2.0
  
  STRIKE_DURATION = 100
  
  def __init__(self):
    Monster     .__init__(self, "caraconque")
    AIControlled.__init__(self)
    
  def _init(self):
    AIControlled._init(self)
    Platform    ._init(self)
    Monster     ._init(self)
    self.carapace = soya.Body(self, soya.Model.get("carapace"))
    self.solid = 1
    self.ai_current_action = b""
    
  def get_photo_name(self):
    if self.distance_to(soya.MAIN_LOOP.camera) > 40.0: return None
    if (not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 1.5, 0.0), self.perso)) and (not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 1.5, 0.0), self.carapace)): return None
    return self.__class__.__name__

  def do_physics(self):
    Monster.do_physics(self)
    
    if   (self.x > 145.0) and (self.speed.x > 0.0): self.speed.x = 0.0
    elif (self.x < -25.0) and (self.speed.x < 0.0): self.speed.x = 0.0
    if   (self.z > 145.0) and (self.speed.z > 0.0): self.speed.z = 0.0
    elif (self.z < -25.0) and (self.speed.z < 0.0): self.speed.z = 0.0
    
  def ai_strike(self):
    self.send_action(ACTION_STRIKE)
    for i in range(75): yield
    for character in self.characters_on:
      if isinstance(character, Hitable):
        character.set_life(character.life - 0.1)
        
    for i in range(35): yield
    self.send_action(ACTION_WAIT)
    yield 1
    
  def ai(self):
    while 1:
      if self.characters_on:
        character = tuple(self.characters_on)[0]
        point.clone(character)
        point.convert_to(self)
        if abs(point.x) > abs(point.z):
          if   point.x < -0.2: 
            new_action = ACTION_TURN_LEFT
          elif point.x >  0.2: 
            new_action = ACTION_TURN_RIGHT
          else:
            new_action = ACTION_WAIT
        else:
          if   point.z < -0.2: 
            new_action = ACTION_ADVANCE
          else:
            if self.ai_current_action != ACTION_WAIT:
              yield self.ai_strike()
            new_action = ACTION_WAIT
              
        if self.ai_current_action != new_action:
          self.send_action(new_action)
          self.ai_current_action = new_action
          
        yield
          
      else:
        if self.ai_current_action != ACTION_WAIT:
          self.send_action(ACTION_WAIT)
          self.ai_current_action = ACTION_WAIT
          
        yield self.ai_wait(10)
      
            
cerealizer.register(Caraconque)




class Echassien(AIControlled, Striking, Monster, Photographiable):
  # We need RADIUS * sqrt(2)/2 < max speed
  RADIUS          = 1.0
  RADIUS_Y        = 1.0
  RADIUS_RAY      = max(RADIUS, 0.1 + RADIUS_Y)
  LOOK_DURATION   = 30
  LOOK_DISTANCE   = 25.0
  LOOK_ANGLE      = 80.0
  CALM_DISTANCE   = 40.0
  FLY             = 9999
  
  STRIKE_DURATION = STRIKE2_DURATION = 22
  
  def __init__(self):
    self.ai_mode = AI_MODE_NORMAL
    Monster     .__init__(self, "echassien")
    AIControlled.__init__(self)
    
  def __getstate__(self): return (Monster.__getstate__(self), self.ai_mode)
  
  def __setstate__(self, state):
    super_state, self.ai_mode = state
    Monster.__setstate__(self, super_state)
    
  def _init(self):
    AIControlled._init(self)
    Monster     ._init(self)
    self.solid = 0
    
    self.right_hand = soya.World(self)
    self.perso.attach_to_bone(self.right_hand, "mainD")
    
    self.weapon_body = soya.Body(self.right_hand, soya.Model.get("epee"))
    self.weapon_body.rotate_x(-90.0)
    self.weapon_body.turn_z(180.0)
    self.weapon_body.set_xyz(0.05, 0.08, -0.08)
    
    self.kill_point = KillPoint(self.weapon_body, 0.0, 0.0, -1.0, 0.5, 0.3)
    
  def get_photo_name(self):
    if self.distance_to(soya.MAIN_LOOP.camera) > 25.0: return None
    if (not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 1.5, 0.0), self.perso)): return None
    return self.__class__.__name__
  
  def can_goto(self, point, min_delta_y = -3.0):
    r = self.level.raypick(point, down, -1.0, 1, 1, raypick_collision, raypick_wall_normal)
    if r:
      raypick_collision  .convert_to(self.level)
      raypick_wall_normal.convert_to(self.level)
      if (raypick_collision.y > -17.5) and (raypick_collision.y - self.y > min_delta_y) and (raypick_wall_normal.y > 0.5):
        return raypick_collision
        
  def ai(self):
    use_strike_2 = 0
    while 1:
      if self.ai_mode == AI_MODE_NORMAL:
        yield self.ai_wait(50)
        
        point.__init__(self.level, self.x + random.uniform(-12.0, 12.0), self.y + 3.0, self.z + random.uniform(-12.0, 12.0))
        if self.can_goto(point):
          yield self.ai_goto(raypick_collision.x, raypick_collision.y, raypick_collision.z)
          
        yield self.ai_look()
        
      else:
        balazar = self.nearest_balazar()
        if (not balazar) or (self.distance_to(balazar) > self.CALM_DISTANCE):
          self.ai_mode = AI_MODE_NORMAL
          continue

        if self.distance_to(balazar) > 11.0:
          use_strike_2 = 0
          point.clone(self)
          vector.set_start_end(self, balazar)
          vector.set_length(9.0)
          point.add_vector(vector)
        else:
          point.clone(balazar)
          vector.set_start_end(balazar, self)
          vector.set_length(2.0)
          point.add_vector(vector)

        point.y += 2.0
        if self.can_goto(point, -20.0):
          yield self.ai_goto(point.x, point.y - 2.0, point.z)
          
        if self.distance_to(balazar) < 3.0:
          yield self.ai_look_toward(balazar)
          if use_strike_2:
            yield self.ai_strike()
            use_strike_2 = 0
          else:
            yield self.ai_strike2()
            use_strike_2 = 1
            
      yield self.ai_wait(1)
      
      
cerealizer.register(Echassien)


class EchassienPriest(AIControlled, Monster, Photographiable):
  # We need RADIUS * sqrt(2)/2 < max speed
  RADIUS          = 1.0
  RADIUS_Y        = 1.0
  RADIUS_RAY      = max(RADIUS, 0.1 + RADIUS_Y)
  LOOK_DURATION   = 30
  LOOK_DISTANCE   = 10.0
  LOOK_ANGLE      = 360.0
  CALM_DISTANCE   = 35.0
  FLY             = 9999
  
  def __init__(self):
    self.ai_mode = AI_MODE_NORMAL
    Monster     .__init__(self, "pretre")
    AIControlled.__init__(self)
    
  def __getstate__(self): return (Monster.__getstate__(self), self.ai_mode)
  
  def __setstate__(self, state):
    super_state, self.ai_mode = state
    Monster.__setstate__(self, super_state)
    
  def _init(self):
    AIControlled._init(self)
    Monster     ._init(self)
    self.solid = 0
    
  def get_photo_name(self):
    if self.distance_to(soya.MAIN_LOOP.camera) > 12.0: return None
    if (not soya.MAIN_LOOP.camera.is_point_visible(soya.Point(self, 0.0, 1.5, 0.0), self.perso)): return None
    return self.__class__.__name__
  
  def can_goto(self, point, min_delta_y = -3.0):
    r = self.level.raypick(point, down, -1.0, 1, 1, raypick_collision, raypick_wall_normal)
    if r:
      raypick_collision  .convert_to(self.level)
      raypick_wall_normal.convert_to(self.level)
      if (raypick_collision.y > -17.5) and (raypick_collision.y - self.y > min_delta_y) and (raypick_wall_normal.y > 0.5):
        return raypick_collision
      
  def ai(self):
    while 1:
      if self.ai_mode == AI_MODE_NORMAL:
        yield self.ai_wait(50)
        
        point.__init__(self.level, self.x + random.uniform(-12.0, 12.0), self.y + 3.0, self.z + random.uniform(-12.0, 12.0))
        if self.can_goto(point):
          yield self.ai_goto(raypick_collision.x, raypick_collision.y, raypick_collision.z)
          
        yield self.ai_look()
        
      else:
        balazar = self.nearest_balazar()
        if (not balazar) or (self.distance_to(balazar) > self.CALM_DISTANCE):
          self.ai_mode = AI_MODE_NORMAL
          yield self.ai_goto(83.2164611816, 23.1427993774, 38.5031585693)
          continue
        
        for i in range(7):
          if i < 5:
            vector.set_start_end(balazar, self)
            vector.set_length(random.uniform(5.0, 20.0))
          else:
            vector.set_start_end(self, balazar)
            vector.__imul__(1.1)
          vector.convert_to(self.level)
          point.__init__(self.level, self.x + vector.x + random.uniform(-10.0, 10.0), self.y + 3.0, self.z + vector.z + random.uniform(-10.0, 10.0))
          if self.can_goto(point, -35.0):
            raypick_collision.convert_to(self.level)
            yield self.ai_goto(raypick_collision.x, raypick_collision.y, raypick_collision.z)
            break
          
        else:
          yield self.ai_wait(10)
      
      
cerealizer.register(EchassienPriest)
