# -*- coding: utf-8 -*-
# Balazar safari photo
# Copyright (C) 2010-2014 Jean-Baptiste LAMY
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, struct, math, random, copy, os
import cerealizer

import soya3 as soya

import balazar_safari_photo, balazar_safari_photo.globdef as globdef
import balazar_safari_photo.tofu.tofu4soya as tofu
from balazar_safari_photo.tofu.sides import *
from balazar_safari_photo.base       import *

ACTION_FALL   = b"f"
ACTION_ASCEND = b"a"
ACTION_GROW   = b"+"

class FallingMushroom(Platform, tofu.Mobile, Dynamic):
  def __init__(self, max_fall = -7.0):
    tofu.Mobile.__init__(self)
    self.max_fall       = max_fall
    self.current_action = ACTION_WAIT
    self.delta_y        = 0.0
    self._init()
    
  def _init(self):
    Platform._init(self)
    self.set_model(soya.Model.get("champignon@pourri"))
    
  def loaded(self):
    tofu.Mobile.loaded(self)
    self._init()
    
  def __getstate__(self):
    return (self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.delta_y, self.max_fall, self.current_action)
  
  def __setstate__(self, state):
    soya.World.__init__(self)
    self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.delta_y, self.max_fall, self.current_action = state
    self.last_noticeable_round = 0
    self.last_important_round  = 0
    
  @side("single", "server")
  def character_on(self, character):
    self.do_action(ACTION_FALL)
    return Platform.character_on(self, character)
    
  def do_action(self, action):
    self.current_action = action
    self.set_current_state_importance(2)
    
  def get_network_state(self):
    return self.current_action + struct.pack("!f", self.y)
  
  def read_network_state(self, f):
    action = f.read(1)
    self.y = struct.unpack("!f", f.read(4))
    self.do_action(action)
    
  @side("single", "server")
  def do_physics(self):
    if self.delta_y <= self.max_fall:
      if self.characters_on:
        if self.current_action == ACTION_FALL: self.do_action(ACTION_WAIT)
      else:
        self.do_action(ACTION_ASCEND)
    if self.current_action == ACTION_ASCEND:
      if self.delta_y >= 0.0: self.do_action(ACTION_WAIT)
      
  def advance_time(self, proportion):
    tofu.Mobile.advance_time(self, proportion)
    if   self.current_action == ACTION_FALL  : self.y -= proportion * 0.3 ; self.delta_y -= proportion * 0.3
    elif self.current_action == ACTION_ASCEND: self.y += proportion * 0.1 ; self.delta_y += proportion * 0.1
    


class TranpolinMushroom(Platform, tofu.Mobile, Dynamic):
  def __init__(self):
    tofu.Mobile.__init__(self)
    self.current_action = ACTION_WAIT
    self.delta_y        = 0.0
    self._init()
    
  def _init(self):
    Platform._init(self)
    self.set_model(soya.Model.get("champignon@bleu"))
    
  def loaded(self):
    tofu.Mobile.loaded(self)
    self._init()
    
  def __getstate__(self):
    return (self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.delta_y, self.current_action)
  
  def __setstate__(self, state):
    soya.World.__init__(self)
    self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.delta_y, self.current_action = state
    self.last_noticeable_round = 0
    self.last_important_round  = 0
    #self._init()
    
  @side("single", "server")
  def character_on(self, character):
    self.do_action(ACTION_ASCEND + character.dumpsuid())
    return Platform.character_on(self, character)
    
  def do_action(self, action):
    self.current_action = action[:1]
    if self.current_action == ACTION_ASCEND:
      character = tofu.Unique.undumpsuid(action[1:])
      character.speed.y = 1.2
      character.jumping = 1
    self.set_current_state_importance(2)
    
  def get_network_state(self):
    return self.current_action + struct.pack("!f", self.y)
  
  def read_network_state(self, f):
    action = f.read(1)
    self.y = struct.unpack("!f", f.read(4))
    self.do_action(action)
    
  @side("single", "server")
  def do_physics(self):
    if self.delta_y >= 8.0:
      if self.characters_on:
        if self.current_action == ACTION_ASCEND: self.do_action(ACTION_WAIT)
      else:
        self.do_action(ACTION_FALL)
    if self.current_action == ACTION_FALL:
      if self.delta_y <= 0.0: self.do_action(ACTION_WAIT)
      
  def advance_time(self, proportion):
    tofu.Mobile.advance_time(self, proportion)
    if   self.current_action == ACTION_FALL  : self.y -= proportion * 0.1; self.delta_y -= proportion * 0.1
    elif self.current_action == ACTION_ASCEND: self.y += proportion * 1.2; self.delta_y += proportion * 1.2
    



GROWING_MUSHROOM_MODELS = None
def load_models():
  global GROWING_MUSHROOM_MODELS
  GROWING_MUSHROOM_MODELS = {
    b"0" : soya.Model        .get("champignon2@pos1"), # 0 appearing
    b"1" : soya.Model        .get("champignon2@pos1"), # 1 young
    b"2" : soya.AnimatedModel.get("champignon2"     ), # 2 growing 1
    b"3" : soya.Model        .get("champignon2@pos2"), # 3 adult
    b"4" : soya.AnimatedModel.get("champignon2"     ), # 4 growing 2
    b"5" : soya.Model        .get("champignon2@pos3"), # 5 old
    b"6" : soya.Model        .get("champignon2@pos3"), # 6 disappearing
    b"7" : None                                      , # 7 dead
  }


class GrowingMushroom(Platform, Hitable, tofu.Mobile, Dynamic):
  RADIUS    = 3.0
  def __init__(self, orig_pos = None, grow_speed = 0.005):
    tofu.Mobile.__init__(self)
    self.orig_pos         = orig_pos
    self.grow             = 0.9
    self.grow_speed       = grow_speed
    self.orig_grow_speed  = grow_speed
    self.sub_mushrooms    = set()
    self.parent_mushroom  = None
    self.current_action   = b"7"
    self.current_rotation = 0
    self.life             = 0.02
    self._init()
    
  def _init(self):
    Platform._init(self)
    self.last_action = ACTION_WAIT
    self.rotate_y(self.current_rotation)
    self.center = soya.Point(self, 0.0, 0.5, 0.0)
    self.set_scale()
    self.set_action()
    #self.add_mini_shader(soya.MiniShader.get("cellshading"))
    
  def hit(self): pass
  
  def set_action(self):
    self.set_model(GROWING_MUSHROOM_MODELS[self.current_action[:1]])
    if   self.current_action == b"2": self.animate_blend_cycle("pousse1", 1.0, 0.2)
    elif self.current_action == b"4": self.animate_blend_cycle("pousse2", 1.0, 0.2)
    
  def set_scale(self):
    if   self.grow_speed < 0.0:
      f = self.grow + 0.0001
      self.set_scale_factors(f, f, f)
      
    elif self.current_action == b"0":
      f = self.grow + 0.0001
      self.set_scale_factors(f, f, f)
    elif self.current_action == b"6":
      f = 1.0001 - self.grow
      self.set_scale_factors(f, f, f)
      
  def __getstate__(self):
    return (self.uid, self.player_name, self.local, self.bot, self.life, self.level, self.x, self.y, self.z, self.current_rotation, self.current_action, self.grow, self.grow_speed, self.orig_grow_speed, self.sub_mushrooms, self.parent_mushroom, self.orig_pos)
  
  def __setstate__(self, state):
    soya.World.__init__(self)
    self.uid, self.player_name, self.local, self.bot, self.life, self.level, self.x, self.y, self.z, self.current_rotation, self.current_action, self.grow, self.grow_speed, self.orig_grow_speed, self.sub_mushrooms, self.parent_mushroom, self.orig_pos = state
    self.last_noticeable_round = 0
    self.last_important_round  = 0
    
  def loaded(self):
    tofu.Mobile.loaded(self)
    self._init()
    
  def generate_actions(self):
    if self.grow >= 1.0:
        next_action = str(int(self.current_action) + 1).encode("utf8")
        if next_action == b"8":
          self.life = 0.02
          point.clone(self.orig_pos)
          point.y += 3.0
          vector.__init__(None, random.uniform(-1.0, 1.0), 0.0, random.uniform(-1.0, 1.0))
          if vector.length() == 0.0: vector.x += 0.1
          vector.set_length(3.0)
          point.add_vector(vector)
          r = self.level.raypick(point, down, -1, 1, 1, raypick_collision, raypick_wall_normal)
          parent_mushroom = None
          if r:
            if isinstance(raypick_collision.parent, GrowingMushroom):
              parent_mushroom = raypick_collision.parent
              if (parent_mushroom.current_action == b"2") or (parent_mushroom.current_action == b"4"):
                # Parent is playing an animation => stop and retry next round
                return
            raypick_collision.convert_to(self.level)
            point.y = raypick_collision.y
          next_action = b"0" + struct.pack("!ffff", point.x, point.y, point.z, random.uniform(0.0, 360.0)) + tofu.Unique.dumpsuid_or_none(parent_mushroom)
        self.do_action(next_action)
    elif self.grow < 0.0:
      self.do_action(b"7")
      
  def character_on(self, character):
    spore = Spore(self.level.parent)
    spore.move(character); spore.y -= 0.5
    if (self.current_action == b"7") or (self.grow_speed < 0.0): return 0
    if (self.current_action == b"6"):
      self.do_action(ACTION_DIE)
      return 0
    return Platform.character_on(self, character)
  
  @side("single", "server")
  def character_off(self, character):
    if not self.sub_mushrooms:
      self.do_action(ACTION_DIE)
    return Platform.character_off(self, character)
  
  def do_action(self, action):
    self.last_action = action
    if action == ACTION_DIE:
      if (not self.sub_mushrooms) and (self.grow_speed >= 0.0):
        self.remove_all_characters()
        if (self.current_action != b"0") and (self.current_action != b"6"):
          self.grow     =  0.99
        self.grow_speed = -0.08
        self.set_current_state_importance(2)
    else:
      self.grow           = 0.0
      self.current_action = action
      self.set_scale_factors(1.0, 1.0, 1.0)
      if   self.current_action.startswith(b"0"):
        self.grow_speed = self.orig_grow_speed
        self.set_identity()
        self.x, self.y, self.z, self.current_rotation = struct.unpack("!ffff", action[1:17])
        self.parent_mushroom = tofu.Unique.undumpsuid(action[17:])
        if self.parent_mushroom: self.parent_mushroom.sub_mushrooms.add(self)
        self.rotate_y(self.current_rotation)
        self.current_action = b"0"
      elif self.current_action == b"6":
        self.remove_all_characters()
      elif self.current_action == b"7":
        spore = BigSpore()
        spore.move(self); spore.y += 1.0
        self.level.add_mobile(spore)
        self.grow_speed = self.orig_grow_speed
        if self.parent_mushroom: self.parent_mushroom.sub_mushrooms.discard(self)
      self.set_action()
      self.set_current_state_importance(2)
      
  def get_network_state(self):
    return self.last_action
  
  def read_network_state(self, f):
    action = f.read(1)
    if action == b"0": action += f.read(18)
    self.do_action(action)
    
  def begin_round(self):
    tofu.Mobile.begin_round(self)
    
  def advance_time(self, proportion):
    if self.grow_speed > 0.0:
      tofu.Mobile.advance_time(self, proportion)
      
      if (self.current_action == b"2") or (self.current_action == b"4"):
        self.grow += 0.01
      elif not self.sub_mushrooms:
        self.grow += self.grow_speed * proportion
        self.set_scale()
        
    else:
      self.grow += self.grow_speed * proportion
      self.set_scale()


#SPORE_MATERIAL = soya.PARTICLE_DEFAULT_MATERIAL
SPORE_MATERIAL = soya.Material.get("x_lumiere_1")

class Spore(soya.Particles, Dynamic):
  def __init__(self, parent = None, nb_particles = 5, life = 5):
    soya.Particles.__init__(self, parent, SPORE_MATERIAL, nb_particles)
    self.auto_generate_particle = 1
    self.removable = 1
    self.set_colors((0.5, 0.95, 0.05, 0.2), (0.5, 0.95, 0.05, 0.7), (0.0, 0.7, 0.0, 0.1))
    self.set_sizes((1.8, 1.8))
    self.life = life
    self.lit = 0
    
  def generate(self, index):
    angle = random.random() * 3.1417
    sx = math.cos(angle)
    sy = math.sqrt(random.random() * 3.0)
    sz = math.sin(angle)
    l = (0.06 * (1.5 + random.random())) / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, 0.5 + random.random(), sx * l, sy * l, sz * l, 0.0, 0.005, 0.0)
    
  def begin_round(self):
    soya.Particles.begin_round(self)
    if   self.life >  0: self.life -= 1
    elif self.life == 0:
      self.auto_generate_particle = 0
      self.life = -1
      
class BigSpore(tofu.Mobile, Dynamic):
  RADIUS = 1.0
  
  def __init__(self):
    tofu.Mobile.__init__(self)
    self.bot  = 1
    self.life = 30
    self._init()
    
  def _init(self):
    Spore(self, 25, 25)
    
  def loaded(self):
    tofu.Mobile.loaded(self)
    self._init()
    
  def __getstate__(self):
    return (self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.life)
  
  def __setstate__(self, state):
    soya.World.__init__(self)
    self.uid, self.player_name, self.local, self.bot, self.level, self.x, self.y, self.z, self.life = state
    
  def generate_actions(self): pass
  
  def do_collisions(self):
    self.life -= 1
    if   self.life >= 0.0:
      for mobile in self.level.mobiles:
        if isinstance(mobile, Hitable) and (mobile.life > 0.0):
          if self.distance_to(mobile) < self.RADIUS + mobile.RADIUS:
            mobile.set_life(mobile.life - 0.3)
    elif self.life < -25:
      self.kill()
      

class Bolet(soya.Body, Photographiable):
  def __init__(self, parent = None):
    soya.Body.__init__(self, parent, soya.Model.get("bolet"))
    
  def get_photo_name(self):
    if self.distance_to(soya.MAIN_LOOP.camera) > 75.0: return None
    for p in [
      soya.Point(self,  0.0,  2.0,  0.0),
      soya.Point(self,  0.0,  6.0,  0.0),
      soya.Point(self,  0.0, 10.0,  0.0),
      soya.Point(self,  0.0, 14.0,  0.0),
      soya.Point(self, -5.0, 14.0,  0.0),
      soya.Point(self,  5.0, 14.0,  0.0),
      soya.Point(self,  0.0, 14.0, -5.0),
      soya.Point(self,  0.0, 14.0,  5.0),
      ]:
      if soya.MAIN_LOOP.camera.is_point_visible(p, self, big = 1):
        return self.__class__.__name__
      
  

cerealizer.register(FallingMushroom)
cerealizer.register(TranpolinMushroom)
cerealizer.register(GrowingMushroom)
cerealizer.register(BigSpore)
cerealizer.register(Bolet)


