# -*- coding: utf-8 -*-
# Balazar safari photo
# Copyright (C) 2010-2014 Jean-Baptiste LAMY
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, struct, math, random, copy, os, atexit
import cerealizer

import soya3 as soya

import balazar_safari_photo, balazar_safari_photo.base, balazar_safari_photo.level, balazar_safari_photo.globdef as globdef
import balazar_safari_photo.flore
import balazar_safari_photo.faune
import balazar_safari_photo.bonus
import balazar_safari_photo.trap
import balazar_safari_photo.weather        as weather
import balazar_safari_photo.interface      as interface
import balazar_safari_photo.tofu           as tofu
import balazar_safari_photo.tofu.tofu4soya as tofu4soya
from balazar_safari_photo.tofu.sides import *


class PlayerID(tofu.PlayerID): pass

class Player(tofu.Player):
  def __init__(self, player_id):
    tofu.Player.__init__(self, player_id)
    
    self.photos       = []
    self.start_pos    = ["champi1", 23.0, -1.0, 108.0]
    #self.start_pos    = ["champi2", 14.7, -3.0, 28.0]
    self.create_character()
    
  def create_character(self):
    character = balazar_safari_photo.base.Balazar()
    level     = character.level = balazar_safari_photo.level.Level.get(self.start_pos[0])
    character.set_xyz(*self.start_pos[1:])
    character.photos = [photo.get_photo() for photo in self.photos]
    
    self.add_mobile(character)
    
  def killed(self):
    self.logout(1)
    
  def login(self, sock, udp_address):
    if not self.mobiles: self.create_character()
      
    self.mobiles[0].photos = [photo.get_photo() for photo in self.photos]
    
    global RESET_LEVEL
    if RESET_LEVEL:
      RESET_LEVEL = False
      self.mobiles[0].life = 1.0
      self.mobiles[0].set_xyz(*self.start_pos[1:])
      self.mobiles[0].chests = { (level_name, chest_id) for (level_name, chest_id) in self.mobiles[0].chests if level_name != self.mobiles[0].level.filename }
      self.mobiles[0].keys[self.mobiles[0].level.filename] = 0
      
    tofu.Player.login(self, sock, udp_address)
    
  def logout(self, save = 1):
    tofu.Player.logout(self, save)
    
  def save_position(self):
    if self.mobiles:
      self.start_pos = [self.mobiles[0].level.filename, self.mobiles[0].x, self.mobiles[0].y, self.mobiles[0].z]
      
  def save(self):
    new_photo_ids = set([photo.id for photo in self.mobiles[0].photos])
    for photo in self.photos[:]:
      if not photo.id in new_photo_ids:
        photo.destroy()
        self.photos.remove(photo)
    old_photo_ids = set([photo.id for photo in self.photos])
    
    for photo in self.mobiles[0].photos:
      if not photo.id in old_photo_ids:
        photo = photo.get_saved_photo(self)
        self.photos.append(photo)
        
    if self.mobiles:
      character_photos = self.mobiles[0].photos
      self.mobiles[0].photos = []
    tofu.Player.save(self)
    if self.mobiles:
      self.mobiles[0].photos = character_photos

    
cerealizer.register(Player, tofu.SavedInAPathHandler(Player))


class MainLoop(tofu4soya.MainLoop):
  def create_scene(self):
    scene        = soya.World()
    self.world   = soya.World(scene)
    self.weather = weather.WeatherAtmosphere(scene)
    self.world.atmosphere = self.weather
    self.camera  = None
    return scene
    
  def __init__(self):
    tofu4soya.MainLoop.__init__(self)
    
    self.music = ""
    if globdef.LIMIT_FPS: self.min_frame_duration = 0.030
    else:                 self.min_frame_duration = 0.0
    
  @side("single", "client")
  def init_interface(self): pass

  def begin_round(self):
    tofu4soya.MainLoop.begin_round(self)
    self.weather.begin_round()
  
  def advance_time(self, proportion):
    tofu4soya.MainLoop.advance_time(self, proportion)
    self.weather.advance_time(proportion)
  
  def render(self):
    soya.MainLoop.render(self)

  def play_sound(self, name):
    if globdef.SOUND_VOLUME > 0:
      soya.SoundPlayer(self.world, soya.Sound.get(name), 0, 0, gain = globdef.SOUND_VOLUME)
      
  def play_music(self, name):
    if human_controller.character:
      human_controller.character.music_name = name
      
    if self.music_name != name:
      self.music_name = name
      if globdef.SOUND_VOLUME > 0 and globdef.MUSIC:
        if self.music: self.world.remove(self.music)
        self.music = soya.SoundPlayer(self.world, soya.Sound.get(name), 1, 0, gain = globdef.SOUND_VOLUME)
        
  def main_loop(self):
    try:
      return tofu4soya.MainLoop.main_loop(self)
    finally:
      if self.music: self.world.remove(self.music)
      


tofu.LOAD_PLAYER_ID = PlayerID.loads
tofu.CREATE_PLAYER  = Player


_SHADERS_IN_MEMORY = []

def init():
  if globdef.SOUND_VOLUME: need_sound = 1
  else:                    need_sound = 0
  soya.init("Balazar safari photo", globdef.SCREEN_WIDTH, globdef.SCREEN_HEIGHT, sound = need_sound, fullscreen = globdef.FULLSCREEN, antialiasing = globdef.QUALITY > 0)
  soya.set_quality(globdef.QUALITY)
  
  soya.AUTO_EXPORTERS_ENABLED = os.path.exists(os.path.join(globdef.APPDIR, ".svn"))
  
  if globdef.QUALITY == 2:
    mini_shader = soya.MiniShader.get("per_pixel_cellshading")
    _SHADERS_IN_MEMORY.append(mini_shader)
    mini_shader.filename = "cellshading"
    soya.MiniShader._alls["cellshading"] = mini_shader
  balazar_safari_photo.flore.load_models()
  

RESET_LEVEL = False

def run_main_loop(action = ["c"]):
  while 1:
    if   action is None:
      break
    
    elif action[0] == "c":
      main_loop = MainLoop()
      if not globdef.LIMIT_FPS: main_loop.min_frame_duration = 0.0
      action = main_loop.main_loop()
      
    elif action[0] == "i":
      inventory = interface.Inventory(action[1])
      action    = inventory.show()
      
    elif action[0] == "r":
      global RESET_LEVEL
      RESET_LEVEL = True
      
      main_loop = MainLoop()
      if not globdef.LIMIT_FPS: main_loop.min_frame_duration = 0.0
      action = main_loop.main_loop()
      
    elif action[0] == "q":
      break
    
  return action
  
def start_single():
  #start_single.reload = 0
  tofu.set_side("single")
  tofu.SAVED_GAME_DIR = globdef.SAVED_GAME_DIR
  tofu.GAME           = globdef.SINGLE_LOGIN
  tofu.PLAYER_IDS     = [PlayerID(globdef.SINGLE_LOGIN, globdef.SINGLE_LOGIN)]
  run_main_loop()


def start_multi():
  if SERVER_PID: # Connect to local server
    tofu.HOST = globdef.SERVER_HOST
    tofu.PORT = globdef.SERVER_PORT
  else:
    tofu.HOST = globdef.CLIENT_HOST
    tofu.PORT = globdef.CLIENT_PORT
    
  tofu.set_side("client")
  tofu.SAVED_GAME_DIR = globdef.SAVED_GAME_DIR
  tofu.GAME           = globdef.CLIENT_LOGIN
  tofu.PLAYER_IDS     = [PlayerID(globdef.CLIENT_LOGIN, globdef.CLIENT_PASSWORD)]
  main_loop = MainLoop()
  if not globdef.LIMIT_FPS: main_loop.min_frame_duration = 0.0
  main_loop.main_loop()
  print(getattr(main_loop, "fps", "-"), "FPS")
  
SERVER_PID = 0

def start_server_spawn():
  global SERVER_PID
  SERVER_PID = os.spawnl(os.P_NOWAIT, sys.executable, sys.executable, sys.argv[0], "--server", globdef.SERVER_HOST, "--port", str(globdef.SERVER_PORT))
  
def start_server():
  tofu.HOST = globdef.SERVER_HOST
  tofu.PORT = globdef.SERVER_PORT
  tofu.GAME = "_server"
  
  tofu.set_side("server")
  tofu.SAVED_GAME_DIR = globdef.SAVED_GAME_DIR
  
  main_loop = MainLoop()
  main_loop.main_loop()
  
def kill_server():
  global SERVER_PID
  if SERVER_PID: 
    import signal
    os.kill(SERVER_PID, signal.SIGINT)
    SERVER_PID = 0
atexit.register(kill_server)

def erase_server_data():
  server_running = SERVER_PID
  if server_running: kill_server()
  
  import shutil
  try:    shutil.rmtree(os.path.join(globdef.SAVED_GAME_DIR, "_server"))
  except: pass
  
  if server_running: start_server_spawn()
  
